# -*- coding: utf-8 -*-
"""Pipeline Runner."""

from __future__ import annotations

import datetime as dt
import importlib.util
import random
import time
from typing import Any, Callable

import humanize
from hamilton import driver
from hamilton.execution import executors
from hamilton.registry import disable_autoload
from hamilton.telemetry import disable_telemetry
from hamilton_sdk.api.clients import UnauthorizedException
from requests.exceptions import ConnectionError, HTTPError

from .. import settings

if importlib.util.find_spec("opentelemetry"):
    from hamilton.plugins import h_opentelemetry

    from ..utils.open_telemetry import init_tracer
else:
    h_opentelemetry = None
    init_tracer = None

if importlib.util.find_spec("mlflow"):
    from hamilton.plugins import h_mlflow
else:
    h_mlflow = None

from hamilton.plugins import h_rich
from hamilton.plugins.h_threadpool import FutureAdapter
from hamilton_sdk.adapters import HamiltonTracker
from hamilton_sdk.tracking import constants
from loguru import logger

if importlib.util.find_spec("distributed"):
    from dask import distributed
    from hamilton.plugins import h_dask
else:
    distributed = None


if importlib.util.find_spec("ray"):
    import ray
    from hamilton.plugins import h_ray
else:
    h_ray = None

from ..cfg import PipelineConfig, ProjectConfig
from ..cfg.pipeline.adapter import AdapterConfig as PipelineAdapterConfig
from ..cfg.pipeline.run import ExecutorConfig, WithAdapterConfig
from ..cfg.project.adapter import AdapterConfig as ProjectAdapterConfig
from ..utils.logging import setup_logging
from .base import load_module

setup_logging(level=settings.LOG_LEVEL)

# from .executor import get_executor


class PipelineRunner:
    """PipelineRunner is responsible for executing a specific pipeline run.
    It handles the loading of the pipeline module, configuration, and execution"""

    def __init__(
        self,
        project_cfg: ProjectConfig,
        pipeline_cfg: PipelineConfig,
    ):
        self.project_cfg = project_cfg
        self.pipeline_cfg = pipeline_cfg
        self.name = pipeline_cfg.name

        if not settings.HAMILTON_TELEMETRY_ENABLED:
            disable_telemetry()
        if not settings.HAMILTON_AUTOLOAD_EXTENSIONS:
            disable_autoload()

    def __enter__(self):
        """Enable use as a context manager."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """No special cleanup required."""
        pass

    def _get_executor(
        self, executor_cfg: str | dict | ExecutorConfig | None = None
    ) -> tuple[executors.BaseExecutor, Callable | None]:
        """
        Get the executor based on the provided configuration.

        Args:
            executor (dict | None): Executor configuration.

        Returns:
            tuple[executors.BaseExecutor, Callable | None]: A tuple containing the executor and shutdown function.
        """
        logger.debug("Setting up executor...")
        if executor_cfg:
            if isinstance(executor_cfg, str):
                executor_cfg = ExecutorConfig(type=executor_cfg)
            elif isinstance(executor_cfg, dict):
                executor_cfg = ExecutorConfig.from_dict(executor_cfg)
            elif not isinstance(executor_cfg, ExecutorConfig):
                raise TypeError(
                    "Executor must be a string, dictionary, or ExecutorConfig instance."
                )

            executor_cfg = self.pipeline_cfg.run.executor.merge(executor_cfg)
        else:
            executor_cfg = self.pipeline_cfg.run.executor

        if executor_cfg.type is None:
            logger.debug(
                "No executor type specified. Using  SynchronousLocalTaskExecutor as default."
            )
            return executors.SynchronousLocalTaskExecutor(), None

        if executor_cfg.type == "threadpool":
            logger.debug(
                f"Using MultiThreadingExecutor with max_workers={executor_cfg.max_workers}"
            )
            return executors.MultiThreadingExecutor(
                max_tasks=executor_cfg.max_workers
            ), None
        elif executor_cfg.type == "processpool":
            logger.debug(
                f"Using MultiProcessingExecutor with max_workers={executor_cfg.max_workers}"
            )
            return executors.MultiProcessingExecutor(
                max_tasks=executor_cfg.max_workers
            ), None
        elif executor_cfg.type == "ray":
            if h_ray:
                logger.debug(
                    f"Using RayTaskExecutor with num_cpus={executor_cfg.num_cpus}"
                )

                return (
                    h_ray.RayTaskExecutor(
                        num_cpus=executor_cfg.num_cpus,
                        ray_init_config=self.project_cfg.adapter.ray.ray_init_config,
                    ),
                    ray.shutdown
                    if self.project_cfg.adapter.ray.shutdown_ray_on_completion
                    else None,
                )
            else:
                logger.warning("Ray is not installed. Using local executor.")
                return executors.SynchronousLocalTaskExecutor(), None
        elif executor_cfg.type == "dask":
            if distributed:
                cluster = distributed.LocalCluster()
                client = distributed.Client(cluster)
                return h_dask.DaskExecutor(client=client), cluster.close
            else:
                logger.warning("Dask is not installed. Using local executor.")
                return executors.SynchronousLocalTaskExecutor(), None
        else:
            logger.warning(
                f"Unknown executor type: {executor_cfg.type}. Using local executor."
            )
            return executors.SynchronousLocalTaskExecutor(), None

    def _get_adapters(
        self,
        with_adapter_cfg: dict | WithAdapterConfig | None = None,
        pipeline_adapter_cfg: dict | PipelineAdapterConfig | None = None,
        project_adapter_cfg: dict | ProjectAdapterConfig | None = None,
        adapter: dict[str, Any] | None = None,
    ) -> list:
        """
        Set the adapters for the pipeline.

        Args:
            with_adapter_cfg (dict | WithAdapterConfig | None): The adapter configuration.
                Overrides the with_adapter settings in the pipeline config.
            pipeline_adapter_cfg (dict | PipelineAdapterConfig | None): The pipeline adapter configuration.
                Overrides the adapter settings in the pipeline config.
            project_adapter_cfg (dict | ProjectAdapterConfig | None): The project adapter configuration.
                Overrides the adapter settings in the project config.
            adapter (dict[str, Any] | None): Any additional hamilton adapters can be passed here.
        """
        logger.debug("Setting up adapters...")
        if with_adapter_cfg:
            if isinstance(with_adapter_cfg, dict):
                with_adapter_cfg = WithAdapterConfig.from_dict(with_adapter_cfg)
            elif not isinstance(with_adapter_cfg, WithAdapterConfig):
                raise TypeError(
                    "with_adapter must be a dictionary or WithAdapterConfig instance."
                )

            with_adapter_cfg = self.pipeline_cfg.run.with_adapter.merge(
                with_adapter_cfg
            )
        else:
            with_adapter_cfg = self.pipeline_cfg.run.with_adapter

        if pipeline_adapter_cfg:
            if isinstance(pipeline_adapter_cfg, dict):
                pipeline_adapter_cfg = PipelineAdapterConfig.from_dict(
                    pipeline_adapter_cfg
                )
            elif not isinstance(pipeline_adapter_cfg, PipelineAdapterConfig):
                raise TypeError(
                    "pipeline_adapter_cfg must be a dictionary or PipelineAdapterConfig instance."
                )

            pipeline_adapter_cfg = self.pipeline_cfg.adapter.merge(pipeline_adapter_cfg)
        else:
            pipeline_adapter_cfg = self.pipeline_cfg.adapter

        if project_adapter_cfg:
            if isinstance(project_adapter_cfg, dict):
                project_adapter_cfg = ProjectAdapterConfig.from_dict(
                    project_adapter_cfg
                )
            elif not isinstance(project_adapter_cfg, ProjectAdapterConfig):
                raise TypeError(
                    "project_adapter_cfg must be a dictionary or ProjectAdapterConfig instance."
                )

            project_adapter_cfg = self.project_cfg.adapter.merge(project_adapter_cfg)
        else:
            project_adapter_cfg = self.project_cfg.adapter

        adapters = []
        if with_adapter_cfg.hamilton_tracker:
            tracker_kwargs = project_adapter_cfg.hamilton_tracker.to_dict()
            tracker_kwargs.update(pipeline_adapter_cfg.hamilton_tracker.to_dict())
            tracker_kwargs["hamilton_api_url"] = tracker_kwargs.pop("api_url", None)
            tracker_kwargs["hamilton_ui_url"] = tracker_kwargs.pop("ui_url", None)

            constants.MAX_DICT_LENGTH_CAPTURE = (
                tracker_kwargs.pop("max_dict_length_capture", None)
                or settings.HAMILTON_MAX_DICT_LENGTH_CAPTURE
            )
            constants.MAX_LIST_LENGTH_CAPTURE = (
                tracker_kwargs.pop("max_list_length_capture", None)
                or settings.HAMILTON_MAX_LIST_LENGTH_CAPTURE
            )
            constants.CAPTURE_DATA_STATISTICS = (
                tracker_kwargs.pop("capture_data_statistics", None)
                or settings.HAMILTON_CAPTURE_DATA_STATISTICS
            )

            tracker = HamiltonTracker(**tracker_kwargs)

            adapters.append(tracker)

        if with_adapter_cfg.mlflow:
            if h_mlflow is None:
                logger.warning("MLFlow is not installed. Skipping MLFlow adapter.")
            else:
                mlflow_kwargs = project_adapter_cfg.mlflow.to_dict()
                mlflow_kwargs.update(pipeline_adapter_cfg.mlflow.to_dict())
                mlflow_adapter = h_mlflow.MLFlowTracker(**mlflow_kwargs)
                adapters.append(mlflow_adapter)

        if with_adapter_cfg.opentelemetry:
            if h_opentelemetry is None:
                logger.warning(
                    "OpenTelemetry is not installed. Skipping OpenTelemetry adapter."
                )
            else:
                otel_kwargs = project_adapter_cfg.opentelemetry.to_dict()
                otel_kwargs.update(pipeline_adapter_cfg.opentelemetry.to_dict())
                trace = init_tracer(**otel_kwargs, name=self.project_cfg.name)
                tracer = trace.get_tracer(self.name)
                otel_adapter = h_opentelemetry.OpenTelemetryTracer(
                    tracer_name=f"{self.project_cfg.name}.{self.name}",
                    tracer=tracer,
                )
                adapters.append(otel_adapter)

        if with_adapter_cfg.progressbar:
            adapters.append(
                h_rich.RichProgressBar(run_desc=f"{self.project_cfg.name}.{self.name}")
            )

        if with_adapter_cfg.future:
            adapters.append(FutureAdapter())

        if with_adapter_cfg.ray:
            if h_ray is None:
                logger.warning("Ray is not installed. Skipping Ray adapter.")
            else:
                ray_kwargs = project_adapter_cfg.ray.to_dict()
                ray_kwargs.update(pipeline_adapter_cfg.ray.to_dict())
                ray_adapter = h_ray.RayGraphAdapter(**ray_kwargs)
                adapters.append(ray_adapter)

        all_adapters = [
            f"{adp}: ✅" if enabled else f"{adp}: ❌"
            for adp, enabled in with_adapter_cfg.to_dict().items()
        ]

        if adapter:
            adapters += list(adapter.values())
            all_adapters += [f"{adp}: ✅" for adp in adapter.keys()]

        logger.debug(f"Adapters enabled: {' | '.join(all_adapters)}")
        return adapters

    def _get_driver(
        self,
        config: dict | None = None,
        cache: bool | dict = False,
        executor_cfg: str | dict | ExecutorConfig | None = None,
        with_adapter_cfg: dict | WithAdapterConfig | None = None,
        pipeline_adapter_cfg: dict | PipelineAdapterConfig | None = None,
        project_adapter_cfg: dict | ProjectAdapterConfig | None = None,
        adapter: dict[str, Any] | None = None,
        reload: bool = False,
    ) -> tuple[driver.Driver, Callable | None]:
        """
        Get the driver and shutdown function for a given pipeline.

        Args:
            config (dict | None): The configuration for the pipeline.
            cache (bool): Use cache or not.
                To fine tune the cache settings, pass a dictionary with the cache settings
                or adjust the pipeline config.
                If set to True, the default cache settings will be used.
            executor_cfg (str | dict | ExecutorConfig | None): The executor to use.
                Overrides the executor settings in the pipeline config.
            with_adapter_cfg (dict | WithAdapterConfig | None): The adapter configuration.
                Overrides the with_adapter settings in the pipeline config.
            pipeline_adapter_cfg (dict | PipelineAdapterConfig | None): The pipeline adapter configuration.
                Overrides the adapter settings in the pipeline config.
            project_adapter_cfg (dict | ProjectAdapterConfig | None): The project adapter configuration.
                Overrides the adapter settings in the project config.
            adapter (dict[str, Any] | None): Any additional Hamilton adapters can be passed here.
            reload (bool): Whether to reload the module.


        Returns:
            tuple[driver.Driver, Callable | None]: A tuple containing the driver and shutdown function.
        """
        logger.debug("Setting up driver...")
        module = load_module(name=self.name, reload=reload)
        executor, shutdown = self._get_executor(executor_cfg)
        adapters = self._get_adapters(
            with_adapter_cfg,
            pipeline_adapter_cfg,
            project_adapter_cfg,
            adapter=adapter,
        )

        config = config or self.pipeline_cfg.run.config

        dr = (
            driver.Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_modules(module)
            .with_config(config)
            .with_local_executor(executors.SynchronousLocalTaskExecutor())
        )

        if cache:
            if isinstance(cache, dict):
                cache = cache or self.pipeline_cfg.run.cache
                dr = dr.with_cache(**cache)
            else:
                dr = dr.with_cache()

        if executor:
            dr = dr.with_remote_executor(executor)

        if adapters:
            dr = dr.with_adapters(*adapters)

        dr = dr.build()
        return dr, shutdown

    def run(
        self,
        inputs: dict | None = None,
        final_vars: list[str] | None = None,
        config: dict | None = None,
        cache: dict | None = None,
        executor_cfg: str | dict | ExecutorConfig | None = None,
        with_adapter_cfg: dict | WithAdapterConfig | None = None,
        pipeline_adapter_cfg: dict | PipelineAdapterConfig | None = None,
        project_adapter_cfg: dict | ProjectAdapterConfig | None = None,
        adapter: dict[str, Any] | None = None,
        reload: bool = False,
        log_level: str | None = None,
        max_retries: int | None = None,
        retry_delay: float | None = None,
        jitter_factor: float | None = None,
        retry_exceptions: tuple = (
            Exception,
            HTTPError,
            UnauthorizedException,
        ),
    ) -> dict[str, Any]:
        """
        Run the pipeline with the given parameters.
        Args:
            inputs (dict | None, optional): The inputs for the pipeline. Defaults to None.
            final_vars (list | None, optional): The final variables for the pipeline. Defaults to None.
            config (dict | None, optional): The config for the hamilton driver. Defaults to None.
            cache (dict | None, optional): The cache configuration. Defaults to None.
            executor_cfg (str | dict | ExecutorConfig | None, optional): The executor to use.
                Overrides the executor settings in the pipeline config. Defaults to None.
            with_adapter_cfg (dict | WithAdapterConfig | None, optional): The adapter configuration.
                Overrides the with_adapter settings in the pipeline config. Defaults to None.
            pipeline_adapter_cfg (dict | PipelineAdapterConfig | None, optional): The pipeline adapter configuration.
                Overrides the adapter settings in the pipeline config. Defaults to None.
            project_adapter_cfg (dict | ProjectAdapterConfig | None, optional): The project adapter configuration.
                Overrides the adapter settings in the project config. Defaults to None.
            adapter (dict[str, Any] | None, optional): Any additional Hamilton adapters can be passed here. Defaults to None.
            reload (bool, optional): Whether to reload the module. Defaults to False.
            log_level (str | None, optional): The log level to use. Defaults to None.
            max_retries (int | None, optional): The maximum number of retry attempts. Defaults to None.
            retry_delay (float | None, optional): The base delay between retries in seconds. Defaults to None.
            jitter_factor (float | None, optional): The factor to apply for jitter. Defaults to None.
            retry_exceptions: tuple | None, optional): The exceptions to catch for retries.
                Defaults to (Exception, HTTPError, UnauthorizedException).

        Returns:
            dict[str, Any]: The result of executing the pipeline.
        """
        self.start_time = dt.datetime.now()

        if log_level or self.pipeline_cfg.run.log_level:
            setup_logging(level=log_level or self.pipeline_cfg.run.log_level)

        logger.info(f"Starting pipeline {self.project_cfg.name}.{self.name}")

        final_vars = final_vars or self.pipeline_cfg.run.final_vars
        inputs = {
            **(self.pipeline_cfg.run.inputs or {}),
            **(inputs or {}),
        }  # <-- inputs override and/or extend config inputs

        max_retries = max_retries or self.pipeline_cfg.run.max_retries
        retry_delay = retry_delay or self.pipeline_cfg.run.retry_delay
        jitter_factor = jitter_factor or self.pipeline_cfg.run.jitter_factor
        retry_exceptions = retry_exceptions or self.pipeline_cfg.run.retry_exceptions

        if not isinstance(retry_exceptions, (tuple, list)):
            retry_exceptions = [retry_exceptions]
        retry_exceptions = [
            eval(exc) if isinstance(exc, str) else exc for exc in retry_exceptions
        ]

        attempts = 1
        last_exception = None

        while attempts <= max_retries:
            logger.debug(f"Attempting to execute pipeline {attempts}/{max_retries}")
            try:
                dr, shutdown = self._get_driver(
                    config=config,
                    cache=cache,
                    executor_cfg=executor_cfg,
                    with_adapter_cfg=with_adapter_cfg,
                    pipeline_adapter_cfg=pipeline_adapter_cfg,
                    project_adapter_cfg=project_adapter_cfg,
                    adapter=adapter,
                    reload=reload,
                )

                res = dr.execute(final_vars=final_vars, inputs=inputs)
                self.end_time = dt.datetime.now()
                self.execution_time = self.end_time - self.start_time
                logger.success(
                    f"Finished: Pipeline {self.project_cfg.name}.{self.name} executed in {humanize.naturaldelta(self.execution_time)}"
                )

                if shutdown is not None:
                    logger.info("Shutting down executor...")
                    shutdown()
                    logger.info("Executor shut down.")

                return res
            except tuple(retry_exceptions) as e:
                # set success to False and handle retries

                if (
                    isinstance(e, HTTPError)
                    or isinstance(e, UnauthorizedException)
                    or isinstance(e, ConnectionError)
                ):
                    if with_adapter_cfg["hamilton_tracker"]:
                        logger.info(
                            "Hamilton Tracker is enabled. Disabling tracker for the next run."
                        )
                        with_adapter_cfg["hamilton_tracker"] = False

                attempts += 1
                last_exception = e

                if attempts <= max_retries:
                    logger.warning(
                        f"Pipeline execution failed (attempt {attempts}/{max_retries}): {e}"
                    )

                    # Calculate base delay with exponential backoff
                    base_delay = retry_delay * (2 ** (attempts - 1))

                    # Add jitter: random value between -jitter_factor and +jitter_factor of the base delay
                    jitter = base_delay * jitter_factor * (2 * random.random() - 1)
                    actual_delay = max(
                        0, base_delay + jitter
                    )  # Ensure non-negative delay

                    logger.debug(
                        f"Retrying in {actual_delay:.2f} seconds (base: {base_delay:.2f}s, jitter: {jitter:.2f}s)"
                    )
                    time.sleep(actual_delay)

                else:
                    # Last attempt failed
                    logger.error(
                        f"Pipeline execution failed after {max_retries} attempts"
                    )
                    raise last_exception


def run_pipeline(
    project_cfg: ProjectConfig,
    pipeline_cfg: PipelineConfig,
    inputs: dict | None = None,
    final_vars: list[str] | None = None,
    config: dict | None = None,
    cache: dict | None = None,
    executor_cfg: str | dict | ExecutorConfig | None = None,
    with_adapter_cfg: dict | WithAdapterConfig | None = None,
    pipeline_adapter_cfg: dict | PipelineAdapterConfig | None = None,
    project_adapter_cfg: dict | ProjectAdapterConfig | None = None,
    adapter: dict[str, Any] | None = None,
    reload: bool = False,
    log_level: str | None = None,
    max_retries: int = 0,
    retry_delay: float = 1.0,
    jitter_factor: float = 0.1,
    retry_exceptions: tuple = (
        Exception,
        HTTPError,
        UnauthorizedException,
    ),  # Adjust to specific exceptions
) -> dict[str, Any]:
    """Run the pipeline with the given parameters.

    Args:

        project_cfg (ProjectConfig): The project configuration.
        pipeline_cfg (PipelineConfig): The pipeline configuration.
        inputs (dict | None, optional): The inputs for the pipeline. Defaults to None.
        final_vars (list | None, optional): The final variables for the pipeline. Defaults to None.
        config (dict | None, optional): The config for the hamilton driver. Defaults to None.
        cache (dict | None, optional): The cache configuration. Defaults to None.
        executor_cfg (str | dict | ExecutorConfig | None, optional): The executor to use.
            Overrides the executor settings in the pipeline config. Defaults to None.
        with_adapter_cfg (dict | WithAdapterConfig | None, optional): The adapter configuration.
            Overrides the with_adapter settings in the pipeline config. Defaults to None.
        pipeline_adapter_cfg (dict | PipelineAdapterConfig | None, optional): The pipeline adapter configuration.
            Overrides the adapter settings in the pipeline config. Defaults to None.
        project_adapter_cfg (dict | ProjectAdapterConfig | None, optional): The project adapter configuration.
            Overrides the adapter settings in the project config. Defaults to None.
        adapter (dict[str, Any] | None, optional): Any additional Hamilton adapters can be passed here. Defaults to None.
        reload (bool, optional): Whether to reload the module. Defaults to False.
        log_level (str | None, optional): The log level to use. Defaults to None.
        max_retries (int, optional): The maximum number of retry attempts. Defaults to 0.
        retry_delay (float, optional): The base delay between retries in seconds. Defaults to 1.0.
        jitter_factor (float, optional): The factor to apply for jitter. Defaults to 0.1.
        retry_exceptions (tuple, optional): A tuple of exception classes to catch for retries. Defaults to (Exception,).

    Returns:

        dict[str, Any]: The result of executing the pipeline.

    Raises:
        Exception: If the pipeline execution fails after the maximum number of retries.
    """

    with PipelineRunner(project_cfg, pipeline_cfg) as runner:
        return runner.run(
            inputs=inputs,
            final_vars=final_vars,
            config=config,
            cache=cache,
            executor_cfg=executor_cfg,
            with_adapter_cfg=with_adapter_cfg,
            pipeline_adapter_cfg=pipeline_adapter_cfg,
            project_adapter_cfg=project_adapter_cfg,
            adapter=adapter,
            reload=reload,
            log_level=log_level,
            max_retries=max_retries,
            retry_delay=retry_delay,
            jitter_factor=jitter_factor,
            retry_exceptions=retry_exceptions,
        )



import functools
import inspect
from typing import Any, Callable, Dict, Tuple

from loguru import logger

from ..settings import LOG_LEVEL
from .logging import setup_logging

# Configure a logger. In a real application, you might get this from a central logging config.
# To see the debug logs from the decorator, set the logging level:
# logging.basicConfig(level=logging.DEBUG)

setup_logging(level=LOG_LEVEL)


def _execute_callback(callback_info: Any, context_exception: Exception = None):
    """
    Helper to execute a callback.
    The callback_info can be a callable, or a tuple (callable, args_tuple, kwargs_dict).
    If context_exception is provided (for on_failure), it can be passed to the callback.
    """
    if not callback_info:
        return

    callback_fn: Callable = None
    cb_args: Tuple = ()
    cb_kwargs: Dict[str, Any] = {}

    is_simple_callable = isinstance(callback_info, Callable)

    if is_simple_callable:
        callback_fn = callback_info
        # For a simple callable in an on_failure context, try to pass the exception.
        if context_exception:
            try:
                sig = inspect.signature(callback_fn)
                if len(sig.parameters) == 1:  # Assumes it takes one positional argument
                    first_param_name = list(sig.parameters.keys())[0]
                    # Avoid passing if it's a **kwargs style param and we have no other indication
                    if sig.parameters[first_param_name].kind in [
                        inspect.Parameter.POSITIONAL_OR_KEYWORD,
                        inspect.Parameter.POSITIONAL_ONLY,
                    ]:
                        cb_args = (context_exception,)
                elif (
                    "exception" in sig.parameters
                ):  # Or if it explicitly takes 'exception' kwarg
                    cb_kwargs["exception"] = context_exception
            except (ValueError, TypeError):  # Some callables might not be inspectable
                logger.debug(
                    f"Could not inspect signature for simple callback {getattr(callback_fn, '__name__', str(callback_fn))}. Exception not passed automatically."
                )

    elif (
        isinstance(callback_info, tuple)
        and len(callback_info) > 0
        and isinstance(callback_info[0], Callable)
    ):
        callback_fn = callback_info[0]

        # Args: callback_info[1]
        if len(callback_info) > 1 and callback_info[1] is not None:
            if isinstance(callback_info[1], tuple):
                cb_args = callback_info[1]
            else:
                logger.warning(
                    f"Callback args for {getattr(callback_fn, '__name__', str(callback_fn))} "
                    f"expected tuple, got {type(callback_info[1])}. Ignoring args."
                )

        # Kwargs: callback_info[2]
        if len(callback_info) > 2 and callback_info[2] is not None:
            if isinstance(callback_info[2], dict):
                cb_kwargs = callback_info[2].copy()  # Use a copy
            else:
                logger.warning(
                    f"Callback kwargs for {getattr(callback_fn, '__name__', str(callback_fn))} "
                    f"expected dict, got {type(callback_info[2])}. Ignoring kwargs."
                )

        # If this is an on_failure call and an exception occurred,
        # pass it if 'exception' kwarg is not set and the callback accepts it.
        if context_exception and "exception" not in cb_kwargs:
            try:
                sig = inspect.signature(callback_fn)
                if "exception" in sig.parameters:
                    cb_kwargs["exception"] = context_exception
            except (ValueError, TypeError):  # Some callables might not be inspectable
                pass  # Cannot inspect, so don't add

    if callback_fn:
        try:
            callback_name = getattr(callback_fn, "__name__", str(callback_fn))
            logger.debug(f"Executing callback: {callback_name}")
            callback_fn(*cb_args, **cb_kwargs)
        except Exception as cb_e:
            callback_name_err = getattr(callback_fn, "__name__", str(callback_fn))
            logger.error(
                f"Error executing callback {callback_name_err}: {cb_e}", exc_info=True
            )
    else:
        logger.warning(
            f"Invalid callback_info format or non-callable provided: {callback_info}"
        )


def run_with_callback(on_success: Any = None, on_failure: Any = None):
    """
    A Python decorator that executes a function within a try-except-finally block,
    allowing for `on_success` and `on_failure` callbacks.

    The `on_success` and `on_failure` arguments can be:
    1. A callable: `my_callback_func`
       - If `on_failure` is a simple callable that accepts one argument or an
         `exception` keyword argument, the caught exception will be passed.
    2. A tuple: `(my_callback_func, (arg1, arg2), {"kwarg1": val1})`
       - The args tuple and kwargs dict are optional (can be `None` or empty).
       - If `on_failure` is used and an exception occurs, the exception instance
         will be passed as a keyword argument `exception` to the failure callback,
         provided the callback accepts it and it's not already in the user-supplied kwargs.

    The decorated function's return value is returned if successful.
    If an exception occurs in the decorated function, it is re-raised after
    the `on_failure` callback is attempted.
    """

    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            success_flag = False
            result_value = None
            caught_exception = None
            try:
                result_value = func(*args, **kwargs)
                success_flag = True
                return result_value
            except Exception as e:
                logger.error(f"Exception in {func.__name__}: {e}", exc_info=True)
                success_flag = False
                caught_exception = e
                # The original exception will be re-raised after the finally block.
                raise
            finally:
                if success_flag:
                    if on_success:
                        # logger.debug(f"Executing on_success callback for {func.__name__}")
                        _execute_callback(on_success)
                else:
                    if on_failure:
                        # logger.debug(f"Executing on_failure callback for {func.__name__}")
                        _execute_callback(
                            on_failure, context_exception=caught_exception
                        )

        return wrapper

    return decorator



# filepath: /Volumes/WD_Blue_1TB/coding/libs/flowerpower/src/flowerpower/worker/rq/concurrent_workers.py
import concurrent.futures
import datetime as dt
import logging
import os
import threading
import time
import traceback
from concurrent.futures import ThreadPoolExecutor

from loguru import logger
from rq import worker
from rq.exceptions import DequeueTimeout
from rq.job import JobStatus
from rq.worker import StopRequested

from flowerpower.utils.logging import setup_logging

utcnow = dt.datetime.utcnow
setup_logging("INFO")


class ThreadWorker(worker.Worker):
    """
    A variation of the RQ Worker that uses a ThreadPoolExecutor to perform
    jobs concurrently within a single worker process.

    Ideal for I/O bound tasks where the GIL is released during waits.
    Jobs share the same memory space within the worker process.
    """

    def __init__(
        self,
        queues,
        name=None,
        max_threads=None,
        default_result_ttl=500,
        connection=None,
        exc_handler=None,
        exception_handlers=None,
        default_worker_ttl=None,
        job_class=None,
        queue_class=None,
        log_job_description=True,
        job_monitoring_interval=30,
        disable_default_exception_handler=False,
        prepare_for_work=True,
        maintenance_interval=600,
    ):
        super().__init__(
            queues,
            name=name,
            default_result_ttl=default_result_ttl,
            connection=connection,
            exc_handler=exc_handler,
            exception_handlers=exception_handlers,
            default_worker_ttl=default_worker_ttl,
            job_class=job_class,
            queue_class=queue_class,
            log_job_description=log_job_description,
            job_monitoring_interval=job_monitoring_interval,
            disable_default_exception_handler=disable_default_exception_handler,
            prepare_for_work=prepare_for_work,
            maintenance_interval=maintenance_interval,
        )

        self.max_threads = (
            max_threads if max_threads is not None else (os.cpu_count() or 1) * 5
        )
        self._executor = None
        self._futures = set()
        self.log = logger

    def work(
        self,
        burst=False,
        logging_level="INFO",
        date_format=worker.DEFAULT_LOGGING_DATE_FORMAT,
        log_format=worker.DEFAULT_LOGGING_FORMAT,
        max_jobs=None,
        with_scheduler=False,
    ):
        """Starts the worker's main loop."""
        self._install_signal_handlers()
        did_perform_work = False
        self.register_birth()
        self.log.info("Worker %s: started, version %s", self.key, worker.VERSION)
        self.set_state(worker.WorkerStatus.STARTED)

        self._executor = ThreadPoolExecutor(max_workers=self.max_threads)
        self._futures = set()
        processed_jobs = 0

        try:
            while True:
                if self._stop_requested or (
                    max_jobs is not None and processed_jobs >= max_jobs
                ):
                    break

                self.run_maintenance_tasks()

                # Wait for space in the thread pool if it's full
                if len(self._futures) >= self.max_threads:
                    done, self._futures = concurrent.futures.wait(
                        self._futures,
                        timeout=1.0,
                        return_when=concurrent.futures.FIRST_COMPLETED,
                    )
                    for future in done:
                        try:
                            future.result()
                        except Exception as e:
                            self.log.error(
                                f"Error in completed job: {e}", exc_info=True
                            )
                    continue

                try:
                    result = self.dequeue_job_and_maintain_ttl(timeout=1)
                except DequeueTimeout:
                    if burst:
                        break
                    time.sleep(0.1)
                    continue
                except StopRequested:
                    break
                except Exception:
                    self.log.error("Error during dequeue:", exc_info=True)
                    time.sleep(1)
                    continue

                if result is None:
                    if burst:
                        did_perform_work = True
                        break
                    time.sleep(0.1)
                    continue

                job, queue = result
                self.log.info("Processing job %s: %s", job.id, job.description)

                try:
                    future = self._executor.submit(self.execute_job, job, queue)
                    self._futures.add(future)
                    future.add_done_callback(
                        lambda f, jid=job.id: self._handle_job_completion(f, jid)
                    )
                except Exception as e:
                    self.log.error(f"Failed to submit job {job.id}: {e}", exc_info=True)
                    continue

                did_perform_work = True
                processed_jobs += 1

        finally:
            if self._executor:
                self._executor.shutdown(wait=True)
            self.register_death()

        return did_perform_work

    def _handle_job_completion(self, future, job_id):
        """Handle completion of a job future, including logging any errors."""
        self._futures.discard(future)
        try:
            future.result()
        except Exception as e:
            self.log.error(f"Error in job {job_id}: {e}", exc_info=True)

    def set_job_status(self, job, status):
        """Sets the job status."""
        if job:
            job.set_status(status)

    def handle_job_success(self, job, queue, started_job_registry):
        """Handles job completion."""
        try:
            if started_job_registry:
                try:
                    started_job_registry.remove(job)
                except NotImplementedError:
                    pass
            job.ended_at = utcnow()
            job.set_status(JobStatus.FINISHED)
        except Exception as e:
            self.log.error(f"Error handling job success for {job.id}: {e}")

    # def handle_job_failure(self, job, queue, started_job_registry, exec_string=None):
    #    """Handles job failure."""
    #    try:
    #        if started_job_registry:
    #            try:
    #                started_job_registry.remove(job)
    #            except NotImplementedError:
    #                pass
    #        job.ended_at = utcnow()
    #        job.set_status(JobStatus.FAILED)
    #    except Exception as e:
    #        self.log.error(f"Error handling job failure for {job.id}: {e}")

    def execute_job(self, job, queue):
        """Execute a job in a worker thread."""
        job_id = job.id if job else "unknown"

        try:
            self.set_job_status(job, JobStatus.STARTED)
            started_job_registry = queue.started_job_registry

            try:
                started_job_registry.add(
                    job,
                    self.job_monitoring_interval * 1000
                    if self.job_monitoring_interval
                    else -1,
                )
            except NotImplementedError:
                pass

            rv = job.perform()
            self.handle_job_success(
                job=job, queue=queue, started_job_registry=started_job_registry
            )
            return rv

        except Exception as e:
            self.log.error(f"Job {job_id} failed: {e}", exc_info=True)
            self.handle_job_failure(
                job=job, queue=queue, started_job_registry=started_job_registry
            )
            raise



import numpy as np
import polars as pl
import pyarrow as pa
import pyarrow.compute as pc

# Pre-compiled regex patterns (identical to original)
INTEGER_REGEX = r"^[-+]?\d+$"
FLOAT_REGEX = r"^[-+]?(?:\d*[.,])?\d+(?:[eE][-+]?\d+)?$"
BOOLEAN_REGEX = r"^(true|false|1|0|yes|ja|no|nein|t|f|y|j|n)$"
BOOLEAN_TRUE_REGEX = r"^(true|1|yes|ja|t|y|j)$"
DATETIME_REGEX = (
    r"^("
    r"\d{4}-\d{2}-\d{2}"  # ISO: 2023-12-31
    r"|"
    r"\d{2}/\d{2}/\d{4}"  # US: 12/31/2023
    r"|"
    r"\d{2}\.\d{2}\.\d{4}"  # German: 31.12.2023
    r"|"
    r"\d{8}"  # Compact: 20231231
    r")"
    r"([ T]\d{2}:\d{2}(:\d{2}(\.\d{1,6})?)?)?"  # Optional time: 23:59[:59[.123456]]
    r"([+-]\d{2}:?\d{2}|Z)?"  # Optional timezone: +01:00, -0500, Z
    r"$"
)

# Float32 range limits
F32_MIN = float(np.finfo(np.float32).min)
F32_MAX = float(np.finfo(np.float32).max)


def unify_schemas(
    schemas: list[pa.Schema], use_large_dtypes: bool = False
) -> pa.Schema:
    """
    Unify a list of PyArrow schemas into a single schema.

    Args:
        schemas (list[pa.Schema]): List of PyArrow schemas to unify.

    Returns:
        pa.Schema: A unified PyArrow schema.
    """
    try:
        return pa.unify_schemas(schemas, promote_options="permissive")
    except (pa.lib.ArrowInvalid, pa.lib.ArrowTypeError) as e:
        _ = e.args[0]
        # If unify_schemas fails, we can try to create a schema with empty tables
        schema = (
            pl.concat(
                [
                    # pl.from_arrow(pa.Table.from_pylist([], schema=schema))
                    pl.from_arrow(schema.empty_table())
                    for schema in schemas
                ],
                how="diagonal_relaxed",
            )
            .to_arrow()
            .schema
        )
        if not use_large_dtypes:
            return convert_large_types_to_normal(schema)
        return schema


def cast_schema(table: pa.Table, schema: pa.Schema) -> pa.Table:
    """
    Cast a PyArrow table to a given schema, updating the schema to match the table's columns.

    Args:
        table (pa.Table): The PyArrow table to cast.
        schema (pa.Schema): The target schema to cast the table to.

    Returns:
        pa.Table: A new PyArrow table with the specified schema.
    """
    # Filter schema fields to only those present in the table
    table_columns = set(table.schema.names)
    filtered_fields = [field for field in schema if field.name in table_columns]
    updated_schema = pa.schema(filtered_fields)
    return table.select(updated_schema.names).cast(updated_schema)


def convert_large_types_to_normal(schema: pa.Schema) -> pa.Schema:
    """
    Convert large types in a PyArrow schema to their standard types.

    Args:
        schema (pa.Schema): The PyArrow schema to convert.

    Returns:
        pa.Schema: A new PyArrow schema with large types converted to standard types.
    """
    # Define mapping of large types to standard types
    type_mapping = {
        pa.large_string(): pa.string(),
        pa.large_binary(): pa.binary(),
        pa.large_utf8(): pa.utf8(),
        pa.large_list(pa.null()): pa.list_(pa.null()),
        pa.large_list_view(pa.null()): pa.list_view(pa.null()),
    }
    # Convert fields
    new_fields = []
    for field in schema:
        field_type = field.type
        # Check if type exists in mapping
        if field_type in type_mapping:
            new_field = pa.field(
                name=field.name,
                type=type_mapping[field_type],
                nullable=field.nullable,
                metadata=field.metadata,
            )
            new_fields.append(new_field)
        # Handle large lists with nested types
        elif isinstance(field_type, pa.LargeListType):
            new_field = pa.field(
                name=field.name,
                type=pa.list_(
                    type_mapping[field_type.value_type]
                    if field_type.value_type in type_mapping
                    else field_type.value_type
                ),
                nullable=field.nullable,
                metadata=field.metadata,
            )
            new_fields.append(new_field)
        # Handle dictionary with large_string, large_utf8, or large_binary values
        elif isinstance(field_type, pa.DictionaryType):
            new_field = pa.field(
                name=field.name,
                type=pa.dictionary(
                    field_type.index_type,
                    type_mapping[field_type.value_type]
                    if field_type.value_type in type_mapping
                    else field_type.value_type,
                    field_type.ordered,
                ),
                # nullable=field.nullable,
                metadata=field.metadata,
            )
            new_fields.append(new_field)
        else:
            new_fields.append(field)

    return pa.schema(new_fields)


def _clean_string_array(array: pa.Array) -> pa.Array:
    """
    Clean string values in a PyArrow array using vectorized operations.
    """
    if len(array) == 0 or array.null_count == len(array):
        return array

    # Trim whitespace using compute functions
    trimmed = pc.utf8_trim_whitespace(array)

    # Create mask for values to convert to null
    empty_mask = pc.equal(trimmed, "")
    dash_mask = pc.equal(trimmed, "-")
    none_mask = pc.equal(trimmed, "None")

    null_mask = pc.or_(pc.or_(empty_mask, dash_mask), none_mask)

    # Apply the mask to set matching values to null
    return pc.if_else(null_mask, None, trimmed)


def _can_downcast_to_float32(array: pa.Array) -> bool:
    """
    Check if float values are within Float32 range using vectorized operations.
    """
    if len(array) == 0 or array.null_count == len(array):
        return True

    # Use compute functions to filter finite values and calculate min/max
    is_finite = pc.is_finite(array)

    # Skip if no finite values
    if not pc.any(is_finite).as_py():
        return True

    # Filter out non-finite values
    finite_array = pc.filter(array, is_finite)

    min_val = pc.min(finite_array).as_py()
    max_val = pc.max(finite_array).as_py()

    return F32_MIN <= min_val <= max_val <= F32_MAX


def _get_optimal_int_type(array: pa.Array) -> pa.DataType:
    """
    Determine the most efficient integer type based on data range.
    """
    # Handle empty or all-null arrays
    if len(array) == 0 or array.null_count == len(array):
        return pa.int8()

    # Use compute functions to get min and max values
    min_max = pc.min_max(array)
    min_val = min_max["min"].as_py()
    max_val = min_max["max"].as_py()

    if min_val >= 0:  # Unsigned
        if max_val <= 255:
            return pa.uint8()
        elif max_val <= 65535:
            return pa.uint16()
        elif max_val <= 4294967295:
            return pa.uint32()
        else:
            return pa.uint64()
    else:  # Signed
        if -128 <= min_val and max_val <= 127:
            return pa.int8()
        elif -32768 <= min_val and max_val <= 32767:
            return pa.int16()
        elif -2147483648 <= min_val and max_val <= 2147483647:
            return pa.int32()
        else:
            return pa.int64()


def _optimize_numeric_array(array: pa.Array, shrink: bool) -> pa.Array:
    """
    Optimize numeric PyArrow array by downcasting when possible.
    Uses vectorized operations for efficiency.
    """
    if not shrink or len(array) == 0 or array.null_count == len(array):
        return array if len(array) > 0 else pa.array([], type=pa.int8())

    # Handle floating point types
    if pa.types.is_floating(array.type):
        if array.type == pa.float64() and _can_downcast_to_float32(array):
            return pc.cast(array, pa.float32())
        return array

    # Handle integer types
    if pa.types.is_integer(array.type):
        # Skip if already optimized to smallest types
        if array.type in [pa.int8(), pa.uint8()]:
            return array

        optimal_type = _get_optimal_int_type(array)
        return pc.cast(array, optimal_type)

    # Default: return unchanged
    return array


def _all_match_regex(array: pa.Array, pattern: str) -> bool:
    """
    Check if all non-null values in array match regex pattern.
    Uses pyarrow.compute.match_substring_regex for vectorized evaluation.
    """
    if len(array) == 0 or array.null_count == len(array):
        return True

    # Check if al values match the pattern
    return pc.all(pc.match_substring_regex(array, pattern, ignore_case=True)).as_py()


def _optimize_string_array(
    array: pa.Array, col_name: str, shrink_numerics: bool, time_zone: str | None = None
) -> pa.Array:
    """
    Convert string PyArrow array to appropriate type based on content analysis.
    Uses fully vectorized operations wherever possible.
    """
    # Handle empty or all-null arrays
    if len(array) == 0:
        return pa.array([], type=pa.int8())
    if array.null_count == len(array):
        return pa.array([None] * len(array), type=pa.int8())

    # Clean string values
    cleaned_array = _clean_string_array(array)

    try:
        # Check for boolean values
        if _all_match_regex(cleaned_array, BOOLEAN_REGEX):
            # Match with TRUE pattern
            true_matches = pc.match_substring_regex(
                array, BOOLEAN_TRUE_REGEX, ignore_case=True
            )

            # Convert to boolean type
            return pc.cast(true_matches, pa.bool_())

        elif _all_match_regex(cleaned_array, INTEGER_REGEX):
            # Convert to integer
            # First replace commas with periods in Polars, then cast
            int_array = pc.cast(
                pc.replace_substring(cleaned_array, ",", "."), pa.int64()
            )

            if shrink_numerics:
                optimal_type = _get_optimal_int_type(int_array)
                return pc.cast(int_array, optimal_type)

            return int_array

        # Check for numeric values
        elif _all_match_regex(cleaned_array, FLOAT_REGEX):
            # Convert to float
            # First replace commas with periods in Polars
            float_array = pc.cast(
                pc.replace_substring(cleaned_array, ",", "."), pa.float64()
            )
            if shrink_numerics and _can_downcast_to_float32(float_array):
                return pc.cast(float_array, pa.float32())

            return float_array

        # Check for datetime values - use polars for conversion as specified
        elif _all_match_regex(cleaned_array, DATETIME_REGEX):
            # Convert via polars

            pl_series = pl.Series(col_name, cleaned_array)
            converted = pl_series.str.to_datetime(
                strict=False, time_unit="us", time_zone=time_zone
            )
            # Convert polars datetime back to pyarrow
            return converted.to_arrow()

    except Exception:
        # Fallback: return cleaned strings on any error
        return cleaned_array

    # Default: return cleaned strings
    return cleaned_array


def _process_column(
    table: pa.Table, col_name: str, shrink_numerics: bool, time_zone: str | None = None
) -> pa.Array:
    """
    Process a single column for type optimization.
    """
    array = table[col_name]

    # Handle all-null columns
    if array.null_count == len(array):
        return pa.array([None] * len(array), type=pa.int8())

    # Process based on current type
    if pa.types.is_floating(array.type) or pa.types.is_integer(array.type):
        return _optimize_numeric_array(array, shrink_numerics)
    elif pa.types.is_string(array.type):
        return _optimize_string_array(array, col_name, shrink_numerics, time_zone)

    # Keep original for other types
    return array


def opt_dtype(
    table: pa.Table,
    include: str | list[str] | None = None,
    exclude: str | list[str] | None = None,
    time_zone: str | None = None,
    shrink_numerics: bool = True,
    strict: bool = False,
) -> pa.Table:
    """
    Optimize data types of a PyArrow Table for performance and memory efficiency.

    This function analyzes each column and converts it to the most appropriate
    data type based on content, handling string-to-type conversions and
    numeric type downcasting. It is the PyArrow equivalent of the Polars
    `opt_dtype` function.

    Args:
        table: PyArrow Table to optimize
        include: Column(s) to include in optimization (default: all columns)
        exclude: Column(s) to exclude from optimization
        time_zone: Optional time zone for datetime parsing
        shrink_numerics: Whether to downcast numeric types when possible
        strict: If True, will raise an error if any column cannot be optimized

    Returns:
        PyArrow Table with optimized data types
    """
    # Normalize include/exclude parameters
    if isinstance(include, str):
        include = [include]
    if isinstance(exclude, str):
        exclude = [exclude]

    # Determine columns to process
    cols_to_process = table.column_names
    if include:
        cols_to_process = [col for col in include if col in table.column_names]
    if exclude:
        cols_to_process = [col for col in cols_to_process if col not in exclude]

    # Process each column and build a new table
    new_columns = []
    for col_name in table.column_names:
        if col_name in cols_to_process:
            try:
                # Process column for optimization
                new_columns.append(
                    _process_column(table, col_name, shrink_numerics, time_zone)
                )
            except Exception as e:
                if strict:
                    raise e
                new_columns.append(table[col_name])
        else:
            new_columns.append(table[col_name])

    # Create a new table with the optimized columns
    return pa.Table.from_arrays(new_columns, names=table.column_names)



import numpy as np
import polars as pl
import polars.selectors as cs

from .datetime import get_timedelta_str, get_timestamp_column

# Pre-compiled regex patterns (identical to original)
INTEGER_REGEX = r"^[-+]?\d+$"
FLOAT_REGEX = r"^[-+]?(?:\d*[.,])?\d+(?:[eE][-+]?\d+)?$"
BOOLEAN_REGEX = r"^(true|false|1|0|yes|ja|no|nein|t|f|y|j|n)$"
BOOLEAN_TRUE_REGEX = r"^(true|1|yes|ja|t|y|j)$"
DATETIME_REGEX = (
    r"^("
    r"\d{4}-\d{2}-\d{2}"  # ISO: 2023-12-31
    r"|"
    r"\d{2}/\d{2}/\d{4}"  # US: 12/31/2023
    r"|"
    r"\d{2}\.\d{2}\.\d{4}"  # German: 31.12.2023
    r"|"
    r"\d{8}"  # Compact: 20231231
    r")"
    r"([ T]\d{2}:\d{2}(:\d{2}(\.\d{1,6})?)?)?"  # Optional time: 23:59[:59[.123456]]
    r"([+-]\d{2}:?\d{2}|Z)?"  # Optional timezone: +01:00, -0500, Z
    r"$"
)

# Float32 range limits
F32_MIN = float(np.finfo(np.float32).min)
F32_MAX = float(np.finfo(np.float32).max)


def _clean_string_expr(col_name: str) -> pl.Expr:
    """Create expression to clean string values."""
    return (
        pl.col(col_name).str.strip_chars().replace({"-": None, "": None, "None": None})
    )


def _can_downcast_to_float32(series: pl.Series) -> bool:
    """Check if float values are within Float32 range."""
    finite_values = series.filter(series.is_finite())
    if finite_values.is_empty():
        return True

    min_val, max_val = finite_values.min(), finite_values.max()
    return F32_MIN <= min_val <= max_val <= F32_MAX


def _optimize_numeric_column(series: pl.Series, col_name: str, shrink: bool) -> pl.Expr:
    """Optimize numeric column types."""
    if not shrink:
        return pl.col(col_name)

    if series.dtype == pl.Float64 and not _can_downcast_to_float32(series):
        return pl.col(col_name)

    return pl.col(col_name).shrink_dtype()


def _optimize_string_column(
    series: pl.Series,
    col_name: str,
    shrink_numerics: bool,
    time_zone: str | None = None,
) -> pl.Expr:
    """Convert string column to appropriate type based on content analysis."""
    # Return early for empty or null-only series
    cleaned_expr = _clean_string_expr(col_name)
    non_null = series.drop_nulls().replace({"-": None, "": None, "None": None})
    if len(non_null) == 0:
        return pl.col(col_name).cast(pl.Int8)

    stripped = non_null.str.strip_chars()
    lowercase = stripped.str.to_lowercase()

    # Check for boolean values
    if lowercase.str.contains(BOOLEAN_REGEX).all():
        return (
            cleaned_expr.str.to_lowercase()
            .str.contains(BOOLEAN_TRUE_REGEX)
            .alias(col_name)
        )

    elif stripped.str.contains(INTEGER_REGEX).all():
        int_expr = cleaned_expr.cast(pl.Int64)
        return (
            int_expr.shrink_dtype().alias(col_name)
            if shrink_numerics
            else int_expr.alias(col_name)
        )

    # Check for numeric values
    elif stripped.str.contains(FLOAT_REGEX).all():
        float_expr = cleaned_expr.str.replace_all(",", ".").cast(pl.Float64)

        if shrink_numerics:
            # Check if values can fit in Float32
            temp_floats = stripped.str.replace_all(",", ".").cast(
                pl.Float64, strict=False
            )
            if _can_downcast_to_float32(temp_floats):
                return float_expr.shrink_dtype().alias(col_name)

        return float_expr.alias(col_name)

    try:
        if stripped.str.contains(DATETIME_REGEX).all():
            return cleaned_expr.str.to_datetime(
                strict=False, time_unit="us", time_zone=time_zone
            ).alias(col_name)
    except pl.exceptions.PolarsError:
        pass

    # Keep original if no conversion applies
    return pl.col(col_name)


def _get_column_expr(
    df: pl.DataFrame, col_name: str, shrink_numerics: bool, time_zone: str | None = None
) -> pl.Expr:
    """Generate optimization expression for a single column."""
    series = df[col_name]

    # Handle all-null columns
    if series.is_null().all():
        return pl.col(col_name).cast(pl.Int8)

    # Process based on current type
    if series.dtype.is_numeric():
        return _optimize_numeric_column(series, col_name, shrink_numerics)
    elif series.dtype == pl.Utf8:
        return _optimize_string_column(series, col_name, shrink_numerics, time_zone)

    # Keep original for other types
    return pl.col(col_name)


def opt_dtype(
    df: pl.DataFrame,
    include: str | list[str] | None = None,
    exclude: str | list[str] | None = None,
    time_zone: str | None = None,
    shrink_numerics: bool = True,
    strict: bool = False,
) -> pl.DataFrame:
    """
    Optimize data types of a Polars DataFrame for performance and memory efficiency.

    This function analyzes each column and converts it to the most appropriate
    data type based on content, handling string-to-type conversions and
    numeric type downcasting.

    Args:
        df: DataFrame to optimize
        include: Column(s) to include in optimization (default: all columns)
        exclude: Column(s) to exclude from optimization
        time_zone: Optional time zone for datetime parsing
        shrink_numerics: Whether to downcast numeric types when possible
        strict: If True, will raise an error if any column cannot be optimized

    Returns:
        DataFrame with optimized data types
    """
    # Normalize include/exclude parameters
    if isinstance(include, str):
        include = [include]
    if isinstance(exclude, str):
        exclude = [exclude]

    # Determine columns to process
    cols_to_process = df.columns
    if include:
        cols_to_process = [col for col in include if col in df.columns]
    if exclude:
        cols_to_process = [col for col in cols_to_process if col not in exclude]

    # Generate optimization expressions for all columns
    expressions = []
    for col_name in cols_to_process:
        try:
            expressions.append(
                _get_column_expr(df, col_name, shrink_numerics, time_zone)
            )
        except Exception as e:
            if strict:
                raise e
            # If strict mode is off, just keep the original column
            continue

    # Apply all transformations at once if any exist
    return df if not expressions else df.with_columns(expressions)


# def opt_dtype(
#     df: pl.DataFrame,
#     include: str | list[str] | None = None,
#     exclude: str | list[str] | None = None,
#     time_zone: str | None = None,
#     shrink_numerics: bool = True,
# ) -> pl.DataFrame:
#     """
#     Analyzes and optimizes the data types of a Polars DataFrame for performance
#     and memory efficiency.

#     This version includes:
#     - Robust numeric, boolean, and datetime casting from strings.
#     - Handling of whitespace and common null-like string values.
#     - Casting of columns containing only nulls to pl.Int8.
#     - Optional shrinking of numeric columns to the smallest possible type.

#     Args:
#         df: The DataFrame to optimize.
#         include: A list of columns to forcefully include in the optimization.
#         exclude: A list of columns to exclude from the optimization.
#         time_zone: Optional time zone for datetime parsing.
#         shrink_numerics: If True, numeric columns (both existing and newly converted from strings)
#             will be downcast to the smallest possible type that can hold their values (e.g., Int64 to Int32, Float64 to Float32),
#             similar to Polars' shrink_dtype() behavior. If False, this shrinking step is skipped.

#     Returns:
#         An optimized Polars DataFrame with improved data types.
#     """
#     # Phase 1: Analysis - Determine columns to process and build a list of
#     # transformation expressions without executing them immediately.
#     if isinstance(include, str):
#         include = [include]
#     if isinstance(exclude, str):
#         exclude = [exclude]

#     cols_to_process = df.columns
#     if include:
#         cols_to_process = [col for col in include if col in df.columns]
#     if exclude:
#         cols_to_process = [col for col in cols_to_process if col not in exclude]

#     expressions = []
#     for col_name in cols_to_process:
#         s = df[col_name]

#         # NEW: If a column is entirely null, cast it to Int8 and skip other checks.
#         if s.is_null().all():
#             expressions.append(pl.col(col_name).cast(pl.Int8))
#             continue

#         dtype = s.dtype

#         # 1. Optimize numeric columns by shrinking their size
#         if dtype.is_numeric():
#             if shrink_numerics:
#                 if dtype == pl.Float64:
#                     column_series = df[col_name]
#                     finite_values_series = column_series.filter(
#                         column_series.is_finite()
#                     )
#                     can_shrink = True
#                     if not finite_values_series.is_empty():
#                         min_finite_val = finite_values_series.min()
#                         max_finite_val = finite_values_series.max()
#                         if (min_finite_val < F32_MIN_FINITE) or (
#                             max_finite_val > F32_MAX_FINITE
#                         ):
#                             can_shrink = False
#                     if can_shrink:
#                         expressions.append(pl.col(col_name).shrink_dtype())
#                     else:
#                         expressions.append(pl.col(col_name))
#                 else:
#                     expressions.append(pl.col(col_name).shrink_dtype())
#             else:
#                 expressions.append(pl.col(col_name))
#             continue

#         # 2. Optimize string columns by casting to more specific types
#         if dtype == pl.Utf8:
#             # Create a cleaned column expression that first strips whitespace, then
#             # replaces common null-like strings.
#             cleaned_col = (
#                 pl.col(col_name)
#                 .str.strip_chars()
#                 .replace({"-": None, "": None, "None": None})
#             )

#             # Analyze a stripped, non-null version of the series to decide the cast type
#             s_non_null = s.drop_nulls()
#             if len(s_non_null) == 0:
#                 # The column only contains nulls or null-like strings.
#                 # Cast to Int8 as requested for all-null columns.
#                 expressions.append(pl.col(col_name).cast(pl.Int8))
#                 continue

#             s_stripped_non_null = s_non_null.str.strip_chars()

#             # Check for boolean type
#             if s_stripped_non_null.str.to_lowercase().str.contains(BOOLEAN_REGEX).all():
#                 expr = cleaned_col.str.to_lowercase().str.contains(BOOLEAN_TRUE_REGEX)
#                 expressions.append(expr.alias(col_name))
#                 continue

#             # Check for numeric type
#             if s_stripped_non_null.str.contains(NUMERIC_REGEX).all():
#                 is_float = s_stripped_non_null.str.contains(r"[.,eE]").any()
#                 numeric_col = cleaned_col.str.replace_all(",", ".")
#                 if is_float:
#                     if shrink_numerics:
#                         temp_float_series = s_stripped_non_null.str.replace_all(
#                             ",", "."
#                         ).cast(pl.Float64, strict=False)
#                         finite_values_series = temp_float_series.filter(
#                             temp_float_series.is_finite()
#                         )
#                         can_shrink = True
#                         if not finite_values_series.is_empty():
#                             min_finite_val = finite_values_series.min()
#                             max_finite_val = finite_values_series.max()
#                             if (min_finite_val < F32_MIN_FINITE) or (
#                                 max_finite_val > F32_MAX_FINITE
#                             ):
#                                 can_shrink = False
#                         base_expr = numeric_col.cast(pl.Float64)
#                         if can_shrink:
#                             expressions.append(base_expr.shrink_dtype().alias(col_name))
#                         else:
#                             expressions.append(base_expr.alias(col_name))
#                     else:
#                         expressions.append(numeric_col.cast(pl.Float64).alias(col_name))
#                 else:
#                     if shrink_numerics:
#                         expressions.append(
#                             numeric_col.cast(pl.Int64).shrink_dtype().alias(col_name)
#                         )
#                     else:
#                         expressions.append(numeric_col.cast(pl.Int64).alias(col_name))
#                 continue

#             # Check for datetime type using a fast heuristic
#             try:
#                 if s_stripped_non_null.str.contains(DATETIME_REGEX).all():
#                     expressions.append(
#                         cleaned_col.str.to_datetime(
#                             strict=False, time_unit="us", time_zone=time_zone
#                         ).alias(col_name)
#                     )
#                     continue
#             except pl.exceptions.PolarsError:
#                 pass

#     # Phase 2: Execution - If any optimizations were identified, apply them
#     # all at once for maximum parallelism and performance.
#     if not expressions:
#         return df

#     return df.with_columns(expressions)


def unnest_all(df: pl.DataFrame, seperator="_", fields: list[str] | None = None):
    def _unnest_all(struct_columns):
        if fields is not None:
            return (
                df.with_columns([
                    pl.col(col).struct.rename_fields([
                        f"{col}{seperator}{field_name}"
                        for field_name in df[col].struct.fields
                    ])
                    for col in struct_columns
                ])
                .unnest(struct_columns)
                .select(
                    list(set(df.columns) - set(struct_columns))
                    + sorted([
                        f"{col}{seperator}{field_name}"
                        for field_name in fields
                        for col in struct_columns
                    ])
                )
            )

        return df.with_columns([
            pl.col(col).struct.rename_fields([
                f"{col}{seperator}{field_name}" for field_name in df[col].struct.fields
            ])
            for col in struct_columns
        ]).unnest(struct_columns)

    struct_columns = [col for col in df.columns if df[col].dtype == pl.Struct]  # noqa: F821
    while len(struct_columns):
        df = _unnest_all(struct_columns=struct_columns)
        struct_columns = [col for col in df.columns if df[col].dtype == pl.Struct]
    return df


def explode_all(df: pl.DataFrame | pl.LazyFrame):
    list_columns = [col for col in df.columns if df[col].dtype == pl.List]
    for col in list_columns:
        df = df.explode(col)
    return df


def with_strftime_columns(
    df: pl.DataFrame | pl.LazyFrame,
    strftime: str | list[str],
    timestamp_column: str = "auto",
    column_names: str | list[str] | None = None,
):
    if timestamp_column is None or timestamp_column == "auto":
        timestamp_column = get_timestamp_column(df)
        if len(timestamp_column):
            timestamp_column = timestamp_column[0]

    if timestamp_column is None:
        raise ValueError("timestamp_column is not specified nor found in the dataframe")

    if isinstance(strftime, str):
        strftime = [strftime]
    if isinstance(column_names, str):
        column_names = [column_names]

    if column_names is None:
        column_names = [
            f"_strftime_{strftime_.replace('%', '').replace('-', '_')}_"
            for strftime_ in strftime
        ]
    # print("timestamp_column, with_strftime_columns", timestamp_column)
    return opt_dtype(
        df.with_columns([
            pl.col(timestamp_column)
            .dt.strftime(strftime_)
            .fill_null(0)
            .alias(column_name)
            for strftime_, column_name in zip(strftime, column_names)
        ]),
        include=column_names,
        strict=False,
    )


def with_truncated_columns(
    df: pl.DataFrame | pl.LazyFrame,
    truncate_by: str | list[str],
    timestamp_column: str = "auto",
    column_names: str | list[str] | None = None,
):
    if timestamp_column is None or timestamp_column == "auto":
        timestamp_column = get_timestamp_column(df)
        if len(timestamp_column):
            timestamp_column = timestamp_column[0]

        if timestamp_column is None:
            raise ValueError(
                "timestamp_column is not specified nor found in the dataframe"
            )
    if isinstance(truncate_by, str):
        truncate_by = [truncate_by]

    if isinstance(column_names, str):
        column_names = [column_names]

    if column_names is None:
        column_names = [
            f"_truncated_{truncate_.replace(' ', '_')}_" for truncate_ in truncate_by
        ]

    truncate_by = [
        get_timedelta_str(truncate_, to="polars") for truncate_ in truncate_by
    ]
    return df.with_columns([
        pl.col(timestamp_column).dt.truncate(truncate_).alias(column_name)
        for truncate_, column_name in zip(truncate_by, column_names)
    ])


def with_datepart_columns(
    df: pl.DataFrame | pl.LazyFrame,
    timestamp_column: str = "auto",
    year: bool = False,
    month: bool = False,
    week: bool = False,
    yearday: bool = False,
    monthday: bool = False,
    day: bool = False,
    weekday: bool = False,
    hour: bool = False,
    minute: bool = False,
    strftime: str | None = None,
):
    if strftime:
        if isinstance(strftime, str):
            strftime = [strftime]
        column_names = [
            f"_strftime_{strftime_.replace('%', '').replace('-', '_')}_"
            for strftime_ in strftime
        ]
    else:
        strftime = []
        column_names = []

    if year:
        strftime.append("%Y")
        column_names.append("year")
    if month:
        strftime.append("%m")
        column_names.append("month")
    if week:
        strftime.append("%W")
        column_names.append("week")
    if yearday:
        strftime.append("%j")
        column_names.append("year_day")
    if monthday:
        strftime.append("%d")
        column_names.append("day")
    if day:
        strftime.append("%d")
        column_names.append("day")
    if weekday:
        strftime.append("%a")
        column_names.append("week_day")
    if hour:
        strftime.append("%H")
        column_names.append("hour")
    if minute:
        strftime.append("%M")
        column_names.append("minute")

    column_names = [col for col in column_names if col not in df.columns]
    # print("timestamp_column, with_datepart_columns", timestamp_column)
    return with_strftime_columns(
        df=df,
        timestamp_column=timestamp_column,
        strftime=strftime,
        column_names=column_names,
    )


def with_row_count(
    df: pl.DataFrame | pl.LazyFrame,
    over: str | list[str] | None = None,
):
    if over:
        if len(over) == 0:
            over = None

    if isinstance(over, str):
        over = [over]

    if over:
        return df.with_columns(pl.lit(1).alias("row_nr")).with_columns(
            pl.col("row_nr").cum_sum().over(over)
        )
    else:
        return df.with_columns(pl.lit(1).alias("row_nr")).with_columns(
            pl.col("row_nr").cum_sum()
        )


# def delta(
#     df1: pl.DataFrame | pl.LazyFrame,
#     df2: pl.DataFrame | pl.LazyFrame,
#     subset: str | list[str] | None = None,
#     eager: bool = False,
# ) -> pl.LazyFrame:
#     columns = sorted(set(df1.columns) & set(df2.columns))

#     if subset is None:
#         subset = columns
#     if isinstance(subset, str):
#         subset = [subset]

#     subset = sorted(set(columns) & set(subset))

#     if isinstance(df1, pl.LazyFrame) and isinstance(df2, pl.DataFrame):
#         df2 = df2.lazy()

#     elif isinstance(df1, pl.DataFrame) and isinstance(df2, pl.LazyFrame):
#         df1 = df1.lazy()

#     df = (
#         pl.concat(
#             [
#                 df1.select(columns)
#                 .with_columns(pl.lit(1).alias("df"))
#                 .with_row_count(),
#                 df2.select(columns)
#                 .with_columns(pl.lit(2).alias("df"))
#                 .with_row_count(),
#             ],
#             how="vertical_relaxed",
#         )
#         .filter((pl.count().over(subset) == 1) & (pl.col("df") == 1))
#         .select(pl.exclude(["df", "row_nr"]))
#     )

#     if eager and isinstance(df, pl.LazyFrame):
#         return df.collect()
#     return df


def drop_null_columns(df: pl.DataFrame | pl.LazyFrame) -> pl.DataFrame | pl.LazyFrame:
    """Remove columns with all null values from the DataFrame."""
    return df.select([col for col in df.columns if df[col].null_count() < df.height])


def unify_schemas(dfs: list[pl.DataFrame | pl.LazyFrame]) -> pl.Schema:
    df = pl.concat([df.lazy() for df in dfs], how="diagonal_relaxed")
    if isinstance(df, pl.LazyFrame):
        return df.collect_schema()
    return df.schema


def cast_relaxed(
    df: pl.DataFrame | pl.LazyFrame, schema: pl.Schema
) -> pl.DataFrame | pl.LazyFrame:
    if isinstance(df, pl.LazyFrame):
        columns = df.collect_schema().names()
    else:
        columns = df.schema.names()
    new_columns = [col for col in schema.names() if col not in columns]
    if len(new_columns):
        return df.with_columns([
            pl.lit(None).alias(new_col) for new_col in new_columns
        ]).cast(schema)
    return df.cast(schema)


def delta(
    df1: pl.DataFrame | pl.LazyFrame,
    df2: pl.DataFrame | pl.LazyFrame,
    subset: list[str] | None = None,
    eager: bool = False,
) -> pl.DataFrame:
    s1 = df1.select(~cs.by_dtype(pl.Null())).collect_schema()
    s2 = df2.select(~cs.by_dtype(pl.Null())).collect_schema()

    columns = sorted(set(s1.names()) & set(s2.names()))

    if subset is None:
        subset = df1.columns
    if isinstance(subset, str):
        subset = [subset]

    subset = sorted(set(columns) & set(subset))

    if isinstance(df1, pl.LazyFrame) and isinstance(df2, pl.DataFrame):
        df2 = df2.lazy()

    elif isinstance(df1, pl.DataFrame) and isinstance(df2, pl.LazyFrame):
        df1 = df1.lazy()

    # cast to equal schema
    unified_schema = unify_schemas([df1.select(subset), df2.select(subset)])

    df1 = df1.cast_relaxed(unified_schema)
    df2 = df2.cast_relaxed(unified_schema)

    df = df1.join(df2, on=subset, how="anti", join_nulls=True)

    if eager and isinstance(df, pl.LazyFrame):
        return df.collect()

    return df


def partition_by(
    df: pl.DataFrame | pl.LazyFrame,
    timestamp_column: str | None = None,
    columns: str | list[str] | None = None,
    strftime: str | list[str] | None = None,
    timedelta: str | list[str] | None = None,
    num_rows: int | None = None,
) -> list[tuple[dict, pl.DataFrame | pl.LazyFrame]]:
    if columns is not None:
        if isinstance(columns, str):
            columns = [columns]
        columns_ = columns.copy()
    else:
        columns_ = []

    drop_columns = columns_.copy()

    if strftime is not None:
        if isinstance(strftime, str):
            strftime = [strftime]

        df = df.with_strftime_columns(
            timestamp_column=timestamp_column, strftime=strftime
        )
        strftime_columns = [
            f"_strftime_{strftime_.replaace('%', '')}_" for strftime_ in strftime
        ]
        columns_ += strftime_columns
        drop_columns += strftime_columns

    if timedelta is not None:
        if isinstance(timedelta, str):
            timedelta = [timedelta]

        df = df.with_duration_columns(
            timestamp_column=timestamp_column, timedelta=timedelta
        )
        timedelta_columns = [f"_timedelta_{timedelta_}_" for timedelta_ in timedelta]
        columns_ += timedelta_columns
        drop_columns += timedelta_columns

    if columns_:
        # datetime_columns = {
        #     col: col in [col.lower() for col in columns_]
        #     for col in [
        #         "year",
        #         "month",
        #         "week",
        #         "yearday",
        #         "monthday",
        #         "weekday",
        #         "strftime",
        #     ]
        #     if col not in [table_col.lower() for table_col in df.columns]
        # }
        datetime_columns = [
            col.lower()
            for col in columns_
            if col
            in [
                "year",
                "month",
                "week",
                "yearday",
                "monthday",
                "weekday",
                "day",
                "hour",
                "minute",
                "strftime",
            ]
            and col not in df.columns
        ]

        datetime_columns = {
            col: col in datetime_columns
            for col in [
                "year",
                "month",
                "week",
                "yearday",
                "monthday",
                "weekday",
                "day",
                "hour",
                "minute",
                "strftime",
            ]
        }
        if any(datetime_columns.values()):
            df = df.with_datepart_columns(
                timestamp_column=timestamp_column, **datetime_columns
            )

        if isinstance(df, pl.LazyFrame):
            df = df.collect()
        columns_ = [col for col in columns_ if col in df.columns]

    if num_rows is not None:
        df = df.with_row_count_ext(over=columns_).with_columns(
            (pl.col("row_nr") - 1) // num_rows
        )
        columns_ += ["row_nr"]
        drop_columns += ["row_nr"]

    if columns_:
        partitions = [
            (p.select(columns_).unique().to_dicts()[0], p.drop(drop_columns))
            for p in df.partition_by(
                by=columns_,
                as_dict=False,
                maintain_order=True,
            )
        ]

        return partitions

    return [({}, df)]


pl.DataFrame.unnest_all = unnest_all
pl.DataFrame.explode_all = explode_all
pl.DataFrame.opt_dtype = opt_dtype
pl.DataFrame.with_row_count_ext = with_row_count
pl.DataFrame.with_datepart_columns = with_datepart_columns
pl.DataFrame.with_duration_columns = with_truncated_columns
pl.DataFrame.with_strftime_columns = with_strftime_columns
pl.DataFrame.cast_relaxed = cast_relaxed
pl.DataFrame.delta = delta
pl.DataFrame.partition_by_ext = partition_by
pl.DataFrame.drop_null_columns = drop_null_columns

pl.LazyFrame.unnest_all = unnest_all
pl.LazyFrame.explode_all = explode_all
pl.LazyFrame.opt_dtype = opt_dtype
pl.LazyFrame.with_row_count_ext = with_row_count
pl.LazyFrame.with_datepart_columns = with_datepart_columns
pl.LazyFrame.with_duration_columns = with_truncated_columns
pl.LazyFrame.with_strftime_columns = with_strftime_columns
pl.LazyFrame.delta = delta
pl.LazyFrame.cast_relaxed = cast_relaxed
pl.LazyFrame.partition_by_ext = partition_by
pl.LazyFrame.drop_null_columns = drop_null_columns



import os

# HAMILTON
HAMILTON_MAX_LIST_LENGTH_CAPTURE = int(
    os.getenv("HAMILTON_MAX_LIST_LENGTH_CAPTURE", 50)
)
HAMILTON_MAX_DICT_LENGTH_CAPTURE = int(
    os.getenv("HAMILTON_MAX_DICT_LENGTH_CAPTURE", 10)
)
HAMILTON_CAPTURE_DATA_STATISTICS = bool(
    os.getenv("HAMILTON_CAPTURE_DATA_STATISTICS", True)
)

HAMILTON_AUTOLOAD_EXTENSIONS = int(os.getenv("HAMILTON_AUTOLOAD_EXTENSIONS", 0))
HAMILTON_TELEMETRY_ENABLED = bool(os.getenv("HAMILTON_TELEMETRY_ENABLED", False))
HAMILTON_API_URL = os.getenv("HAMILTON_API_URL", "http://localhost:8241")
HAMILTON_UI_URL = os.getenv("HAMILTON_UI_URL", "http://localhost:8242")



import importlib
from typing import Any, Optional

from loguru import logger

from ..cfg.project import ProjectConfig
from ..fs import AbstractFileSystem
from ..utils.logging import setup_logging

if importlib.util.find_spec("apscheduler"):
    from .apscheduler import APSBackend, APSManager
else:
    APSBackend = None
    APSManager = None
if importlib.util.find_spec("rq"):
    from .rq import RQBackend, RQManager
else:
    RQBackend = None
    RQManager = None
from .base import BaseBackend, BaseJobQueueManager

setup_logging()


class JobQueueBackend:
    """A factory class for creating backend instances for different job queue types.

    This class provides a unified interface for creating backend instances that handle
    the storage, queuing, and event management for different job queue types. Each backend
    type provides specific implementations for:
    - Job storage and persistence
    - Queue management
    - Event handling and communication
    - Result storage

    Example:
        ```python
        # Create RQ backend with Redis
        rq_backend = JobQueueBackend(
            job_queue_type="rq",
            uri="redis://localhost:6379/0",
            queues=["high", "default", "low"]
        )

        # Create APScheduler backend with PostgreSQL and Redis
        aps_backend = JobQueueBackend(
            job_queue_type="apscheduler",
            data_store={
                "type": "postgresql",
                "uri": "postgresql+asyncpg://user:pass@localhost/db"
            },
            event_broker={
                "type": "redis",
                "uri": "redis://localhost:6379/0"
            }
        )
        ```
    """

    def __new__(
        cls,
        job_queue_type: str,
        **kwargs,
    ) -> BaseBackend:
        """Create a new backend instance based on the specified job queue type.

        Args:
            job_queue_type: The type of backend to create. Valid values are:
                - "rq": Redis Queue backend using Redis
                - "apscheduler": APScheduler backend supporting various databases
                    and event brokers
            **kwargs: Backend-specific configuration options:
                For RQ:
                    - uri (str): Redis connection URI
                    - queues (list[str]): List of queue names
                    - result_ttl (int): Time to live for results in seconds
                For APScheduler:
                    - data_store (dict): Data store configuration
                    - event_broker (dict): Event broker configuration
                    - cleanup_interval (int): Cleanup interval in seconds
                    - max_concurrent_jobs (int): Maximum concurrent jobs

        Returns:
            BaseBackend: An instance of RQBackend or APSBackend depending on
                the specified job queue type.

        Raises:
            ValueError: If an invalid job queue type is specified.
            RuntimeError: If backend initialization fails due to configuration
                or connection issues.

        Example:
            ```python
            # Create RQ backend
            rq_backend = Backend(
                job_queue_type="rq",
                uri="redis://localhost:6379/0",
                queues=["high", "default", "low"],
                result_ttl=3600
            )

            # Create APScheduler backend with PostgreSQL and Redis
            aps_backend = Backend(
                job_queue_type="apscheduler",
                data_store={
                    "type": "postgresql",
                    "uri": "postgresql+asyncpg://user:pass@localhost/db",
                    "schema": "scheduler"
                },
                event_broker={
                    "type": "redis",
                    "uri": "redis://localhost:6379/0"
                },
                cleanup_interval=300,
                max_concurrent_jobs=10
            )
            ```
        """
        if job_queue_type == "rq" and RQBackend is not None:
            return RQBackend(**kwargs)
        elif job_queue_type == "apscheduler" and APSBackend is not None:
            return APSBackend(**kwargs)
        else:
            if job_queue_type == "rq" and RQBackend is None:
                logger.warning(
                    "RQ is not installed. `JobQueueBackend` is not initialized and using the job queue is disabled. "
                    "Install rq to use RQ. `uv pip install flowerpower[rq]` or `uv add flowerpower[rq]`"
                )
                return None
            elif job_queue_type == "apscheduler" and APSBackend is None:
                logger.warning(
                    "APScheduler is not installed. `JobQueueBackend` is not initialized and using the job queue is disabled. "
                    "Install apscheduler to use APScheduler. `uv pip install flowerpower[apscheduler]` or `uv add flowerpower[apscheduler]`"
                )
                return None
            else:
                raise ValueError(
                    f"Invalid job queue type: {job_queue_type}. Valid types: ['rq', 'apscheduler']"
                )


class JobQueueManager:
    """A factory class for creating job queue instances for job scheduling and execution.

    This class provides a unified interface for creating different types of job queue instances
    (RQ, APScheduler, Huey) based on the specified backend type. Each job queue type provides
    different capabilities for job scheduling and execution.

    The job queue instances handle:
    - Job scheduling and execution
    - Background task processing
    - Job queue management
    - Result storage and retrieval

    Example:
        ```python
        # Create an RQ job queue
        rq_worker = JobQueueManager(
            type="rq",
            name="my_worker",
            log_level="DEBUG"
        )

        # Create an APScheduler job queue with custom backend
        from flowerpower.job_queue.apscheduler import APSBackend
        backend_config = APSBackend(
            data_store={"type": "postgresql", "uri": "postgresql+asyncpg://user:pass@localhost/db"},
            event_broker={"type": "redis", "uri": "redis://localhost:6379/0"}
        )
        aps_worker = JobQueueManager(
            type="apscheduler",
            name="scheduler",
            backend=backend_config
        )

        ```
    """

    def __new__(
        cls,
        type: str | None = None,
        name: str | None = None,
        base_dir: str | None = ".",
        backend: JobQueueBackend | None = None,
        storage_options: Optional[dict[str, Any]] = None,
        fs: AbstractFileSystem | None = None,
        log_level: str | None = None,
        **kwargs,
    ) -> BaseJobQueueManager:
        """Create a new job queue instance based on the specified backend type.

        Args:
            type: The type of job queue to create. Valid values are:
                - "rq": Redis Queue job queue for Redis-based job queuing
                - "apscheduler": APScheduler job queue for advanced job scheduling
            name: Name of the job queue instance. Used for identification in logs
                and monitoring.
            base_dir: Base directory for job queue files and configuration. Defaults
                to current working directory if not specified.
            backend: Pre-configured backend instance. If not provided, one will
                be created based on configuration settings.
            storage_options: Options for configuring filesystem storage access.
                Example: {"mode": "async", "root": "/tmp", "protocol": "s3"}
            fs: Custom filesystem implementation for storage operations.
                Example: S3FileSystem, LocalFileSystem, etc.
            log_level: Logging level for the job queue. Valid values are:
                "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"
            **kwargs: Additional configuration options passed to the specific
                job queue implementation.

        Returns:
            BaseJobQueueManager: An instance of the specified job queue type (RQManager,
                APSManager).

        Raises:
            ValueError: If an invalid job queue type is specified.
            ImportError: If required dependencies for the chosen job queue type
                are not installed.
            RuntimeError: If job queue initialization fails due to configuration
                or connection issues.

        Example:
            ```python
            # Basic RQ job queue
            worker = JobQueueManager(type="rq", name="basic_worker")

            # APScheduler with custom logging and storage
            worker = JobQueueManager(
                type="apscheduler",
                name="scheduler",
                base_dir="/app/data",
                storage_options={"mode": "async"},
                log_level="DEBUG"
            )

            ```
        """
        if type is None:
            type = ProjectConfig.load(
                base_dir=base_dir,
                name=name,
                fs=fs,
                storage_options=storage_options or {},
            ).job_queue.type

        if type == "rq":
            if RQManager is not None:
                return RQManager(
                    name=name,
                    base_dir=base_dir,
                    backend=backend,
                    storage_options=storage_options,
                    fs=fs,
                    log_level=log_level,
                    **kwargs,
                )
            else:
                logger.warning(
                    "`JobQueueManager` can not be initialized. This might be due to missing dependencies (RQ), invalid configuration or backend not being available."
                )
                return None

        elif type == "apscheduler":
            if APSManager is not None:
                return APSManager(
                    name=name,
                    base_dir=base_dir,
                    backend=backend,
                    storage_options=storage_options,
                    fs=fs,
                    log_level=log_level,
                    **kwargs,
                )
            else:
                logger.warning(
                    "`JobQueueManager` can not be initialized. This might be due to missing dependencies (APScheduler), invalid configuration or backend not being available."
                )
                return None

        else:
            raise ImportError(
                f"Invalid job queue type: {type}. Valid types: ['rq', 'apscheduler']"
            )


__all__ = [
    "JobQueueManager",
    "RQManager",
    "APSManager",
    # "HueyWorker",
    "JobQueueBackend",
    "RQBackend",
    "APSBackend",
]



import re
from typing import Any

import pyarrow as pa
import pyarrow.compute as pc
from sqlglot import exp, parse_one

from .datetime import timestamp_from_string
from .polars import pl

# Compile regex patterns once for efficiency
SPLIT_PATTERN = re.compile(
    r"<=|<|>=|>|=|!=|\s+[n,N][o,O][t,T]\s+[i,I][n,N]\s+|\s+[i,I][n,N]\s+|"
    r"\s+[i,I][s,S]\s+[n,N][o,O][t,T]\s+[n,N][u,U][l,L]{2}\s+|\s+[i,I][s,S]\s+[n,N][u,U][l,L]{2}\s+"
)
LOGICAL_OPERATORS_PATTERN = re.compile(
    r"\s+[a,A][n,N][d,D] [n,N][o,O][t,T]\s+|\s+[a,A][n,N][d,D]\s+|"
    r"\s+[o,O][r,R] [n,N][o,O][t,T]\s+|\s+[o,O][r,R]\s+"
)


def sql2pyarrow_filter(string: str, schema: pa.Schema) -> pc.Expression:
    """
    Generates a filter expression for PyArrow based on a given string and schema.

    Parameters:
        string (str): The string containing the filter expression.
        schema (pa.Schema): The PyArrow schema used to validate the filter expression.

    Returns:
        pc.Expression: The generated filter expression.

    Raises:
        ValueError: If the input string is invalid or contains unsupported operations.
    """

    def parse_value(val: str, type_: pa.DataType) -> Any:
        """Parse and convert value based on the field type."""
        if isinstance(val, (tuple, list)):
            return type(val)(parse_value(v, type_) for v in val)

        if pa.types.is_timestamp(type_):
            return timestamp_from_string(val, exact=False, tz=type_.tz)
        elif pa.types.is_date(type_):
            return timestamp_from_string(val, exact=True).date()
        elif pa.types.is_time(type_):
            return timestamp_from_string(val, exact=True).time()

        elif pa.types.is_integer(type_):
            return int(float(val.strip("'").replace(",", ".")))
        elif pa.types.is_floating(type_):
            return float(val.strip("'").replace(",", "."))
        elif pa.types.is_boolean(type_):
            return val.lower().strip("'") in ("true", "1", "yes")
        else:
            return val.strip("'")

    def _parse_part(part: str) -> pc.Expression:
        match = SPLIT_PATTERN.search(part)
        if not match:
            raise ValueError(f"Invalid condition: {part}")

        sign = match.group().lower().strip()
        field, val = [p.strip() for p in SPLIT_PATTERN.split(part)]

        if field not in schema.names:
            raise ValueError(f"Unknown field: {field}")

        type_ = schema.field(field).type
        val = parse_value(val, type_)

        operations = {
            ">=": lambda f, v: pc.field(f) >= v,
            ">": lambda f, v: pc.field(f) > v,
            "<=": lambda f, v: pc.field(f) <= v,
            "<": lambda f, v: pc.field(f) < v,
            "=": lambda f, v: pc.field(f) == v,
            "!=": lambda f, v: pc.field(f) != v,
            "in": lambda f, v: pc.field(f).isin(v),
            "not in": lambda f, v: ~pc.field(f).isin(v),
            "is null": lambda f, v: pc.field(f).is_null(nan_is_null=True),
            "is not null": lambda f, v: ~pc.field(f).is_null(nan_is_null=True),
        }

        if sign not in operations:
            raise ValueError(f"Unsupported operation: {sign}")

        return operations[sign](field, val)

    parts = LOGICAL_OPERATORS_PATTERN.split(string)
    operators = [op.lower().strip() for op in LOGICAL_OPERATORS_PATTERN.findall(string)]

    if len(parts) == 1:
        return _parse_part(parts[0])

    expr = _parse_part(parts[0])
    for part, operator in zip(parts[1:], operators):
        if operator == "and":
            expr = expr & _parse_part(part)
        elif operator == "and not":
            expr = expr & ~_parse_part(part)
        elif operator == "or":
            expr = expr | _parse_part(part)
        elif operator == "or not":
            expr = expr | ~_parse_part(part)
        else:
            raise ValueError(f"Unsupported logical operator: {operator}")

    return expr


def sql2polars_filter(string: str, schema: pl.Schema) -> pl.Expr:
    """
    Generates a filter expression for Polars based on a given string and schema.

    Parameters:
        string (str): The string containing the filter expression.
        schema (pl.Schema): The Polars schema used to validate the filter expression.

    Returns:
        pl.Expr: The generated filter expression.

    Raises:
        ValueError: If the input string is invalid or contains unsupported operations.
    """

    def parse_value(val: str, dtype: pl.DataType) -> Any:
        """Parse and convert value based on the field type."""
        if isinstance(val, (tuple, list)):
            return type(val)(parse_value(v, dtype) for v in val)

        if dtype == pl.Datetime:
            return timestamp_from_string(val, exact=False, tz=dtype.time_zone)
        elif dtype == pl.Date:
            return timestamp_from_string(val, exact=True).date()
        elif dtype == pl.Time:
            return timestamp_from_string(val, exact=True).time()
        elif dtype in (pl.Int8, pl.Int16, pl.Int32, pl.Int64):
            return int(float(val.strip("'").replace(",", ".")))
        elif dtype in (pl.Float32, pl.Float64):
            return float(val.strip("'").replace(",", "."))
        elif dtype == pl.Boolean:
            return val.lower().strip("'") in ("true", "1", "yes")
        else:
            return val.strip("'")

    def _parse_part(part: str) -> pl.Expr:
        match = SPLIT_PATTERN.search(part)
        if not match:
            raise ValueError(f"Invalid condition: {part}")

        sign = match.group().lower().strip()
        field, val = [p.strip() for p in SPLIT_PATTERN.split(part)]

        if field not in schema.names():
            raise ValueError(f"Unknown field: {field}")

        dtype = schema[field]
        val = parse_value(val, dtype)

        operations = {
            ">=": lambda f, v: pl.col(f) >= v,
            ">": lambda f, v: pl.col(f) > v,
            "<=": lambda f, v: pl.col(f) <= v,
            "<": lambda f, v: pl.col(f) < v,
            "=": lambda f, v: pl.col(f) == v,
            "!=": lambda f, v: pl.col(f) != v,
            "in": lambda f, v: pl.col(f).is_in(v),
            "not in": lambda f, v: ~pl.col(f).is_in(v),
            "is null": lambda f, v: pl.col(f).is_null(),
            "is not null": lambda f, v: pl.col(f).is_not_null(),
        }

        if sign not in operations:
            raise ValueError(f"Unsupported operation: {sign}")

        return operations[sign](field, val)

    parts = LOGICAL_OPERATORS_PATTERN.split(string)
    operators = [op.lower().strip() for op in LOGICAL_OPERATORS_PATTERN.findall(string)]

    if len(parts) == 1:
        return _parse_part(parts[0])

    expr = _parse_part(parts[0])
    for part, operator in zip(parts[1:], operators):
        if operator == "and":
            expr = expr & _parse_part(part)
        elif operator == "and not":
            expr = expr & ~_parse_part(part)
        elif operator == "or":
            expr = expr | _parse_part(part)
        elif operator == "or not":
            expr = expr | ~_parse_part(part)
        else:
            raise ValueError(f"Unsupported logical operator: {operator}")

    return expr


def get_table_names(sql_query):
    return [table.name for table in parse_one(sql_query).find_all(exp.Table)]



from .csv import CSVDatasetReader, CSVFileReader
from .deltatable import DeltaTableReader
from .duckdb import DuckDBReader
from .json import JsonDatasetReader, JsonFileReader
from .mssql import MSSQLReader
from .mysql import MySQLReader
from .oracle import OracleDBReader
from .parquet import ParquetDatasetReader, ParquetFileReader
from .postgres import PostgreSQLReader
from .pydala import PydalaDatasetReader
from .sqlite import SQLiteReader

__all__ = [
    "CSVFileReader",
    "CSVDatasetReader",
    "DeltaTableReader",
    "DuckDBReader",
    "JsonFileReader",
    "JsonDatasetReader",
    "MSSQLReader",
    "MySQLReader",
    "OracleDBReader",
    "ParquetFileReader",
    "ParquetDatasetReader",
    "PostgreSQLReader",
    "PydalaDatasetReader",
    "SQLiteReader",
]



from ...cfg.base import BaseConfig


class MqttConfig(BaseConfig):
    username: str | None = None
    password: str | None = None
    host: str | None = "localhost"
    port: int | None = 1883
    topic: str | None = None
    first_reconnect_delay: int = 1
    max_reconnect_count: int = 5
    reconnect_rate: int = 2
    max_reconnect_delay: int = 60
    transport: str = "tcp"
    clean_session: bool = True
    client_id: str | None = None
    client_id_suffix: str | None = None



import datetime as dt
import importlib
import os

import duckdb
import pandas as pd
import polars as pl
import pyarrow as pa
import pyarrow.dataset as pds
from deltalake import DeltaTable
from fsspec import AbstractFileSystem

from ...fs.ext import path_to_glob


def get_serializable_schema(
    data: (
        pd.DataFrame
        | pl.DataFrame
        | pl.LazyFrame
        | duckdb.DuckDBPyRelation
        | pa.Table
        | pa.Schema
        | pa.RecordBatch
        | pa.RecordBatchReader
        | pds.Dataset
    ),
) -> dict[str, str]:
    """
    Convert DataFrame dtypes to a serializable dictionary.

    Args:
        data:  DataFrame

    Returns:
        dict mapping column names to dtype strings
    """
    if isinstance(data, pd.DataFrame):
        return {col: str(dtype) for col, dtype in data.dtypes.items()}
    elif isinstance(data, pl.DataFrame):
        return data.schema.to_python()
    elif isinstance(data, pl.LazyFrame):
        return data.collect_schema().to_python()
    elif isinstance(data, duckdb.DuckDBPyRelation):
        return dict(zip(data.columns, [str(dtype) for dtype in data.types]))
    elif isinstance(
        data, pa.Table | pa.RecordBatch | pa.RecordBatchReader | pds.Dataset
    ):
        return dict(zip(data.schema.names, [str(dtype) for dtype in data.schema.types]))
    elif isinstance(data, pa.Schema):
        return dict(zip(data.names, [str(dtype) for dtype in data.types]))


def get_dataframe_metadata(
    df: pd.DataFrame
    | pl.DataFrame
    | pl.LazyFrame
    | pa.Table
    | pa.RecordBatch
    | pa.RecordBatchReader
    | list[
        pd.DataFrame
        | pl.DataFrame
        | pl.LazyFrame
        | pa.Table
        | pa.RecordBatch
        | pa.RecordBatchReader
    ],
    path: str | list[str] | None = None,
    format: str | None = None,
    topic: str | None = None,
    num_files: int | None = None,
    partition_columns: list[str] | None = None,
    fs: AbstractFileSystem | None = None,
    **kwargs,
) -> dict:
    """
    Get metadata for a DataFrame.

    Args:
        df: DataFrame
        path: Path to the file(s) that the DataFrame was loaded from
        fs: Optional filesystem
        kwargs: Additional metadata fields

    Returns:
        dict: DataFrame metadata
    """
    if isinstance(df, list):
        schema = get_serializable_schema(df[0])
        num_rows = sum(df_.shape[0] for df_ in df)
    else:
        schema = get_serializable_schema(df)
        num_rows = df.shape[0] if hasattr(df, "shape") else None

    if path is not None and num_files is None:
        if isinstance(path, list):
            num_files = len(path)
        else:
            path_ = path_to_glob(path=path, format=format)
            num_files = len(fs.glob(path_)) if fs is not None else None

    if partition_columns is not None:
        schema = {k: v for k, v in schema.items() if k not in partition_columns}

    metadata = {
        "path": path,
        "topic": topic,
        "format": format,
        "timestamp": int(dt.datetime.now().timestamp() * 1000),
        "schema": schema,
        "partition_columns": partition_columns,
        "num_columns": len(schema),
        "num_rows": num_rows,
        "num_files": num_files,
    }
    metadata.update(kwargs)
    return {k: v for k, v in metadata.items() if v is not None}


def get_duckdb_metadata(
    rel: duckdb.DuckDBPyRelation,
    path: str,
    format: str,
    fs: AbstractFileSystem | None = None,
    include_shape: bool = False,
    include_num_files: bool = False,
    partition_columns: list[str] | None = None,
    **kwargs,
) -> dict:
    """
    Get metadata for a DuckDBPyRelation.

    Args:
        rel: DuckDBPyRelation
        path: Path to the file(s) that the DuckDBPyRelation was loaded from
        fs: Filesystem
        include_shape: Include shape in metadata
        include_num_files: Include number of files in metadata
        kwargs: Additional metadata fields

    Returns:
        dict: DuckDBPyRelation metadata
    """

    schema = get_serializable_schema(rel)
    if include_shape:
        shape = rel.shape
    else:
        shape = None
    if partition_columns is not None:
        schema = {k: v for k, v in schema.items() if k not in partition_columns}

    metadata = {
        "path": path,
        "format": format,
        "timestamp": dt.datetime.now().timestamp(),
        "schema": schema,
        "partition_columns": partition_columns,
        "num_columns": shape[1] if shape else None,
        "num_rows": shape[0] if shape else None,
        "num_files": len(fs.glob(path)) if include_num_files else None,
    }
    metadata.update(kwargs)
    return {k: v for k, v in metadata.items() if v is not None}


def get_pyarrow_dataset_metadata(
    ds: pds.Dataset,
    path: str,
    format: str,
    **kwargs,
) -> dict:
    schema = get_serializable_schema(ds.schema)
    files = ds.files

    metadata = {
        "path": path or os.path.dirname(files[0]),
        "format": format,
        "timestamp": dt.datetime.now().timestamp(),
        "schema": schema,
        "partition_columns": ds.partitioning.schema.names if ds.partitioning else None,
        "num_columns": len(ds.schema),
        "num_rows": None,
        "num_files": len(files),
    }
    metadata.update(kwargs)
    return metadata


def get_delta_metadata(
    dtable: DeltaTable,
    path: str,
    **kwargs,
) -> dict:
    dt_meta = dtable.metadata()
    dt_schema = dtable.schema().to_pyarrow()
    metadata = {
        "path": path,
        "format": "delta",
        "timestamp": dt.datetime.now().timestamp(),
        "schema": dict(zip(dt_schema.names, [str(x) for x in dt_schema.types])),
        "partition_columns": dt_meta.partition_columns
        if hasattr(dt_meta, "partition_columns")
        else None,
        "num_columns": len(dt_schema),
        "num_files": len(dtable.files()),
        "name": dt_meta.name or kwargs.get("name", None),
        "description": dt_meta.description or kwargs.get("description", None),
        "id": dt_meta.id or kwargs.get("id", None),
    }

    return {k: v for k, v in metadata.items() if v is not None}


if importlib.util.find_spec("orjson"):
    import orjson

    def get_mqtt_metadata(
        payload: bytes | dict[str, any],
        topic: str | None = None,
        **kwargs,
    ) -> dict:
        if isinstance(payload, bytes):
            payload = orjson.loads(payload)

        schema = get_serializable_schema(payload)
        metadata = {
            "topic": topic,
            "format": "mqtt",
            "timestamp": dt.datetime.now().timestamp(),
            "schema": schema,
            "num_columns": len(schema),
            "num_rows": len(payload),
            "name": kwargs.get("name", None),
            "description": kwargs.get("description", None),
            "id": kwargs.get("id", None),
        }
        return metadata

else:

    def get_mqtt_metadata(*args, **kwargs):
        raise ImportError("orjson not installed")



from msgspec import field

from ..base import BaseFileWriter


# @attrs.define
class JsonFileWriter(BaseFileWriter, gc=False):
    """JSON file writer.

    This class is responsible for writing dataframes to JSON files.

    Examples:
        ```python
        writer = JsonFileWriter(df, "data.json")
        writer.write()
        ```
    """

    format: str = field(default="json")


# @attrs.define
class JsonDatasetWriter(BaseFileWriter, gc=False):
    """JSON dataset writer.

    This class is responsible for writing dataframes to JSON dataset.

    Examples:
        ```python
        writer = JsonDatasetWriter([df1, df2], "json_data/")
        writer.write()
        ```

    """

    format: str = field(default="json")



"""
Base scheduler interface for FlowerPower.

This module defines the abstract base classes for scheduling operations
that can be implemented by different backend providers (APScheduler, RQ, etc.).
"""

import abc
import importlib
import os
import posixpath
import sys
import urllib.parse
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any, TypeVar

if importlib.util.find_spec("sqlalchemy"):
    from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
else:
    create_async_engine = None
    AsyncEngine = TypeVar("AsyncEngine")

from ..cfg import ProjectConfig
from ..fs import AbstractFileSystem, get_filesystem
# from ..utils.misc import update_config_from_dict
from ..settings import BACKEND_PROPERTIES, CACHE_DIR, CONFIG_DIR, PIPELINES_DIR


class BackendType(str, Enum):
    POSTGRESQL = "postgresql"
    MYSQL = "mysql"
    SQLITE = "sqlite"
    MONGODB = "mongodb"
    MQTT = "mqtt"
    REDIS = "redis"
    NATS_KV = "nats_kv"
    MEMORY = "memory"

    @property
    def properties(self):
        return BACKEND_PROPERTIES[self.value]

    @property
    def uri_prefix(self) -> str:
        return self.properties.get("uri_prefix", "")

    @property
    def default_port(self):
        return self.properties.get("default_port")

    @property
    def default_host(self) -> str:
        return self.properties.get("default_host", "")

    @property
    def default_username(self) -> str:
        return self.properties.get("default_username", "")

    @property
    def default_password(self) -> str:
        return self.properties.get("default_password", "")

    @property
    def default_database(self) -> str:
        return self.properties.get("default_database", "")

    @property
    def is_sqla_type(self) -> bool:
        return self.properties.get("is_sqla_type", False)

    @property
    def is_mongodb_type(self) -> bool:
        return self.value == "mongodb"

    @property
    def is_mqtt_type(self) -> bool:
        return self.value == "mqtt"

    @property
    def is_redis_type(self) -> bool:
        return self.value == "redis"

    @property
    def is_nats_kv_type(self) -> bool:
        return self.value == "nats_kv"

    @property
    def is_memory_type(self) -> bool:
        return self.value == "memory"

    @property
    def is_sqlite_type(self) -> bool:
        return self.value == "sqlite"

    def gen_uri(
        self,
        host: str | None = None,
        port: int | None = None,
        username: str | None = None,
        password: str | None = None,
        database: str | None = None,
        ssl: bool = False,
        ca_file: str | None = None,
        cert_file: str | None = None,
        key_file: str | None = None,
        verify_ssl: bool = False,
    ) -> str:
        # Handle host and port
        host = host or self.default_host
        port = port or self.default_port
        database = database or self.default_database
        username = username or self.default_username
        password = password or self.default_password

        # components: List[str] = []
        # Get the appropriate URI prefix based on backend type and SSL setting
        if self.is_redis_type:
            uri_prefix = "rediss://" if ssl else "redis://"
        elif self.is_nats_kv_type:
            uri_prefix = "nats+tls://" if ssl else "nats://"
        elif self.is_mqtt_type:
            uri_prefix = "mqtts://" if ssl else "mqtt://"
            if ssl and port == 1883:
                port = 8883
        else:
            uri_prefix = self.uri_prefix

        # Handle authentication
        if username and password:
            auth = f"{urllib.parse.quote(username)}:{urllib.parse.quote(password)}@"
        elif username:
            auth = f"{urllib.parse.quote(username)}@"
        elif password:
            auth = f":{urllib.parse.quote(password)}@"
        else:
            auth = ""

        port_part = f":{port}"  # if port is not None else self.default_port

        # Special handling for SQLite and memory types
        if self.is_sqlite_type or self.is_memory_type:
            if self.is_sqlite_type:
                if database:
                    return f"{uri_prefix}{database}"
                else:
                    return f"{uri_prefix}"
            return "memory://"

        # Build path component
        database = database or self.default_database
        path = f"/{database}" if database else ""

        # Construct base URI
        base_uri = f"{uri_prefix}{auth}{host}{port_part}{path}"

        # Prepare query parameters for SSL files
        query_params: list[str] = []

        if ssl:
            # Always add ssl query parameter if ssl=True
            if self.value == "postgresql":
                query_params.append("ssl=verify-full" if verify_ssl else "ssl=allow")
                if ca_file:
                    query_params.append(f"sslrootcert={urllib.parse.quote(ca_file)}")
                if cert_file:
                    query_params.append(f"sslcert={urllib.parse.quote(cert_file)}")
                if key_file:
                    query_params.append(f"sslkey={urllib.parse.quote(key_file)}")
            elif self.value == "mysql":
                query_params.append("ssl=true")
                if ca_file:
                    query_params.append(f"ssl_ca={urllib.parse.quote(ca_file)}")
                if cert_file:
                    query_params.append(f"ssl_cert={urllib.parse.quote(cert_file)}")
                if key_file:
                    query_params.append(f"ssl_key={urllib.parse.quote(key_file)}")
            elif self.is_mongodb_type:
                query_params.append("tls=true")
                if ca_file:
                    query_params.append(f"tlsCAFile={urllib.parse.quote(ca_file)}")
                if cert_file and key_file:
                    query_params.append(
                        f"tlsCertificateKeyFile={urllib.parse.quote(cert_file)}"
                    )
            elif self.is_redis_type:
                if not verify_ssl:
                    query_params.append("ssl_cert_reqs=none")
                if ca_file:
                    query_params.append(f"ssl_ca_certs={urllib.parse.quote(ca_file)}")
                if cert_file:
                    query_params.append(f"ssl_certfile={urllib.parse.quote(cert_file)}")
                if key_file:
                    query_params.append(f"ssl_keyfile={urllib.parse.quote(key_file)}")
            elif self.is_nats_kv_type:
                query_params.append("tls=true")
                if ca_file:
                    query_params.append(f"tls_ca_file={urllib.parse.quote(ca_file)}")
                if cert_file:
                    query_params.append(
                        f"tls_cert_file={urllib.parse.quote(cert_file)}"
                    )
                if key_file:
                    query_params.append(f"tls_key_file={urllib.parse.quote(key_file)}")
            elif self.is_mqtt_type:
                query_params.append("tls=true")
                if ca_file:
                    query_params.append(f"tls_ca_file={urllib.parse.quote(ca_file)}")
                if cert_file:
                    query_params.append(
                        f"tls_cert_file={urllib.parse.quote(cert_file)}"
                    )
                if key_file:
                    query_params.append(f"tls_key_file={urllib.parse.quote(key_file)}")

        # Compose query string if Any params exist
        query_string = ""
        if query_params:
            query_string = "?" + "&".join(query_params)

        return f"{base_uri}{query_string}"


@dataclass(slots=True)
class BaseBackend:
    type: BackendType | str | None = None
    uri: str | None = None
    username: str | None = None
    password: str | None = None
    host: str | None = None
    port: int | None = None
    database: str | None = None
    ssl: bool = False
    ca_file: str | None = None
    cert_file: str | None = None
    key_file: str | None = None
    verify_ssl: bool = False
    _kwargs: dict = field(default_factory=dict)
    _sqla_engine: AsyncEngine | None = (
        None  # SQLAlchemy async engine instance for SQL backends
    )
    _client: Any | None = None  # Native client instance for non-SQL backends

    def __post_init__(self):
        if self.type is None:
            self.type = "memory"

        elif isinstance(self.type, str):
            try:
                self.type = BackendType[self.type.upper()]
            except KeyError:
                raise ValueError(
                    f"Invalid backend type: {self.type}. Valid types: {[bt.value for bt in BackendType]}"
                )

        if not self.uri:
            self.uri = self.type.gen_uri(
                username=self.username,
                password=self.password,
                host=self.host,
                port=self.port,
                database=self.database,
                ssl=self.ssl,
                ca_file=self.ca_file,
                cert_file=self.cert_file,
                key_file=self.key_file,
                verify_ssl=self.verify_ssl,
            )

        # Setup is handled by backend-specific implementations

    @classmethod
    def from_dict(cls, d: dict) -> "BaseBackend":
        return cls(**d)


class BaseTrigger(abc.ABC):
    """
    Abstract base class for schedule triggers.

    A trigger determines when a scheduled job should be executed.
    """

    def __init__(self, trigger_type: str):
        self.trigger_type = trigger_type

    @abc.abstractmethod
    def get_trigger_instance(self, **kwargs) -> Any:
        """
        Get the backend-specific trigger instance.

        Args:
            **kwargs: Keyword arguments specific to the trigger type

        Returns:
            Any: A backend-specific trigger instance
        """
        pass


class BaseJobQueueManager:
    """
    Abstract base class for scheduler workers (APScheduler, RQ, etc.).
    Defines the required interface for all scheduler backends.

    Can be used as a context manager:

    ```python
    with RQManager(name="test") as manager:
        manager.add_job(job1)
    ```
    """

    def __enter__(self):
        """Context manager entry - returns self for use in with statement."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit - ensures workers are stopped."""
        if hasattr(self, "_worker_process") and self._worker_process is not None:
            self.stop_worker()
        if hasattr(self, "_worker_pool") and self._worker_pool is not None:
            self.stop_worker_pool()
        if hasattr(self, "_worker") and self._worker is not None:
            self.stop_worker()
        if hasattr(self, "_scheduler") and self._scheduler is not None:
            self.stop_scheduler()
        return False  # Don't suppress exceptions

    def __init__(
        self,
        type: str | None = None,
        name: str | None = None,
        base_dir: str | None = None,
        backend: BaseBackend | None = None,
        storage_options: dict = None,
        fs: AbstractFileSystem | None = None,
        **kwargs,
    ):
        """
        Initialize the APScheduler backend.

        Args:
            name: Name of the scheduler
            base_dir: Base directory for the FlowerPower project
            backend: APSBackend instance with data store and event broker
            storage_options: Storage options for filesystem access
            fs: Filesystem to use
            cfg_override: Configuration overrides for the worker
        """
        self.name = name or ""
        self._base_dir = base_dir or str(Path.cwd())
        # self._storage_options = storage_options or {}
        self._backend = backend
        self._type = type
        self._pipelines_dir = kwargs.get("pipelines_dir", PIPELINES_DIR)
        self._cfg_dir = CONFIG_DIR

        if storage_options is not None:
            cached = True
            cache_storage = posixpath.join(
                posixpath.expanduser(CACHE_DIR), self._base_dir.split("://")[-1]
            )
            os.makedirs(cache_storage, exist_ok=True)
        else:
            cached = False
            cache_storage = None
        if not fs:
            fs = get_filesystem(
                self._base_dir,
                storage_options=storage_options,
                cached=cached,
                cache_storage=cache_storage,
            )
        self._fs = fs
        self._storage_options = storage_options or fs.storage_options

        self._add_modules_path()
        self._load_config()

    def _load_config(self) -> None:
        """Load the configuration.

        Args:
            cfg_updates: Configuration updates to apply
        """
        self.cfg = ProjectConfig.load(
            base_dir=self._base_dir, job_queue_type=self._type, fs=self._fs
        ).job_queue

    def _add_modules_path(self):
        """
        Sync the filesystem.

        Returns:
            None
        """
        if self._fs.is_cache_fs:
            self._fs.sync_cache()
            project_path = self._fs._mapper.directory
            modules_path = posixpath.join(project_path, self._pipelines_dir)

        else:
            # Use the base directory directly if not using cache
            project_path = self._fs.path
            modules_path = posixpath.join(project_path, self._pipelines_dir)

        if project_path not in sys.path:
            sys.path.insert(0, project_path)

        if modules_path not in sys.path:
            sys.path.insert(0, modules_path)
